# @file ci_build_plugin
# Plugin that supports adding tests or operations to the ci environment
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Plugin that supports adding tests or operations to the ci environment."""

import logging
import os
from typing import TextIO

from edk2toollib.log.junit_report_format import JunitReportTestCase
from edk2toollib.uefi.edk2.path_utilities import Edk2Path

from edk2toolext.environment.plugin_manager import PluginManager
from edk2toolext.environment.plugintypes.uefi_helper_plugin import HelperFunctions
from edk2toolext.environment.var_dict import VarDict


class ICiBuildPlugin(object):
    """Plugin that supports adding tests or operations to the ci environment."""
    def RunBuildPlugin(
        self,
        packagename: str,
        Edk2pathObj: Edk2Path,
        pkgconfig: dict,
        environment: VarDict,
        PLM: PluginManager,
        PLMHelper: HelperFunctions,
        tc: JunitReportTestCase,
        output_stream: TextIO
    ) -> int:
        """External function of plugin.

        This function is used to perform the task of the CiBuild Plugin

        Args:
            packagename (str): edk2 path to package (workspace/package path relative)
            Edk2pathObj (Edk2Path): Edk2Path configured with workspace and package path
            pkgconfig (dict): Package config
            environment (VarDict): Environment config
            PLM (PluginManager): Plugin manager instance
            PLMHelper (HelperFunctions): Plugin helper object instace
            tc (obj): test case that needs state configured for reporting by plugin
            output_stream (StringIO): output stream from this plugin via logging

        Returns:
            (int): >0 - number of errors found
            (int): 0 - passed successfully
            (int): -1 - skipped for missing prereq
        """

    def GetTestName(self, packagename: str, environment: object) -> tuple[str, str]:
        """Provides the test case and class name.

        Given the package name and configuration provide the caller
        the name of the test case and the class name.  These are both used in logging
        and reporting test status.

        Args:
            packagename (str): Package Name
            environment (EnvDict): Environment Dictionary configuration

        Returns:
            (Tuple[str, str]): (test case name, test case base class name)
        """

    def RunsOnTargetList(self) -> list[str]:
        """Returns a list of edk2 TARGETs that this plugin would like to run on.

        !!! note "Known target values"
            DEBUG, RELEASE, NOOPT, NO-TARGET

        !!! hint
            If the plugin is not Target specific it should return a list of one element of "NO-TARGET"
        """
        return ["NO-TARGET"]

    def WalkDirectoryForExtension(
        self,
        extensionlist: list[str],
        directory: os.PathLike,
        ignorelist: list[str] = None
    ) -> list[os.PathLike]:
        """Walks a file directory recursively for all items ending in certain extension.

        Args:
            extensionlist (List[str]): list of file extensions
            directory (os.PathLike): absolute path to directory to start looking
            ignorelist (List[str]): a list of case insensitive filenames to ignore (Optional)

        Returns:
            (List): file paths to matching files
        """
        if not isinstance(extensionlist, list):
            logging.critical("Expected list but got " + str(type(extensionlist)))
            raise TypeError("extensionlist must be a list")

        if directory is None:
            logging.critical("No directory given")
            raise TypeError("directory is None")

        if not os.path.isabs(directory):
            logging.critical("Directory not abs path")
            raise ValueError("directory is not an absolute path")

        if not os.path.isdir(directory):
            logging.critical("Invalid find directory to walk")
            raise ValueError("directory is not a valid directory path")

        if ignorelist is not None:
            if not isinstance(ignorelist, list):
                logging.critical("Expected list but got " + str(type(ignorelist)))
                raise TypeError("ignorelist must be a list")

            ignorelist_lower = list()
            for item in ignorelist:
                ignorelist_lower.append(item.lower())

        extensionlist_lower = list()
        for item in extensionlist:
            extensionlist_lower.append(item.lower())

        returnlist = list()
        for Root, Dirs, Files in os.walk(directory):
            for File in Files:
                for Extension in extensionlist_lower:
                    if File.lower().endswith(Extension):
                        ignoreIt = False
                        if (ignorelist is not None):
                            for c in ignorelist_lower:
                                if (File.lower().startswith(c)):
                                    ignoreIt = True
                                    break
                        if not ignoreIt:
                            returnlist.append(os.path.join(Root, File))

        return returnlist



## @file
# manage multiple workspace file.
#
# This file is required to make Python interpreter treat the directory
# as containing package.
#
# file slightly modified from Edk2 BaseTools\Source\Python\Common\MultipleWorkspace.py
#
# Copyright (c) 2015 - 2018, Intel Corporation. All rights reserved.<BR>
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
#
"""This file is required to make Python interpreter treat the directory as a containing package.

File is slightly modified from Edk2 BaseTools/Source/Python/Common/MultipleWorkspace.py
"""

import os
import warnings
from typing import Optional


class MultipleWorkspace(object):
    """This class manages multiple workspace behavior.

    Attributes:
        WORKSPACE (str): defined the current workspace
        PACKAGES_PATH (str): defined the other WORKSPACE
    """
    WORKSPACE = ''
    PACKAGES_PATH = None

    @classmethod
    def convertPackagePath(cls: 'MultipleWorkspace', Ws: str, Path: str) -> str:
        """Convert path to match workspace.

        Args:
            cls (MultipleWorkspace): The class pointer
            Ws (str): The current WORKSPACE
            Path (str): Path to be converted to match workspace

        Returns:
            (str): Converted path.
        """
        if str(os.path.normcase(Path)).startswith(Ws):
            return os.path.join(Ws, Path[len(Ws) + 1:])
        return Path

    @classmethod
    def setWs(cls: 'MultipleWorkspace', Ws: str, PackagesPath: Optional[list[str]]=None) -> None:
        """Set WORKSPACE and PACKAGES_PATH environment.

        Args:
            cls (MultipleWorkspace): The class pointer
            Ws (str): initialize WORKSPACE variable
            PackagesPath (str): initialize PackagesPath variable
        """
        cls.WORKSPACE = Ws
        if PackagesPath:
            cls.PACKAGES_PATH = [cls.convertPackagePath(Ws, os.path.normpath(
                Path.strip())) for Path in PackagesPath.split(os.pathsep)]
        else:
            cls.PACKAGES_PATH = []

    @classmethod
    def join(cls: 'MultipleWorkspace', Ws: str, *p: str) -> str:
        """Rewrite os.path.join.

        Args:
            cls (MultipleWorkspace): The class pointer
            Ws (str): the current WORKSPACE
            *p (str): path of the inf/dec/dsc/fdf/conf file

        Returns:
            (str): absolute path of the specified file
        """
        warnings.warn(
            "MultipleWorkspace is deprecated. Use Edk2Path.GetAbsolutePathOnThisSystemFromEdk2RelativePath().",
            DeprecationWarning, stacklevel=2)
        Path = os.path.join(Ws, *p)
        if not os.path.exists(Path):
            for Pkg in cls.PACKAGES_PATH:
                Path = os.path.join(Pkg, *p)
                if os.path.exists(Path):
                    return Path
            Path = os.path.join(Ws, *p)
        return Path

    @classmethod
    def relpath(cls: 'MultipleWorkspace', Path: str, Ws: str) -> str:
        """Rewrite os.path.relpath.

        Args:
            cls (MultipleWorkspace): The class pointer
            Path (str): path of the inf/dec/dsc/fdf/conf file
            Ws (str): the current WORKSPACE

        Returns:
            (str): the relative path of specified file
        """
        warnings.warn(
            "MultipleWorkspace is deprecated. use Edk2Path.GetEdk2RelativePathOnThisSystemFromAbsolutePath().",
            DeprecationWarning, stacklevel=2)
        for Pkg in cls.PACKAGES_PATH:
            if Path.lower().startswith(Pkg.lower()):
                Path = os.path.relpath(Path, Pkg)
                return Path
        if Path.lower().startswith(Ws.lower()):
            Path = os.path.relpath(Path, Ws)
        return Path

    @classmethod
    def getWs(cls: 'MultipleWorkspace', Ws: str, Path: str) -> str:
        """Get valid workspace for the path.

        Args:
            cls (MultipleWorkspace): The class pointer
            Ws (str): the current WORKSPACE
            Path (str): path of the inf/dec/dsc/fdf/conf file

        Returns:
            (str): valid workspace relative to the specified file path
        """
        absPath = os.path.join(Ws, Path)
        if not os.path.exists(absPath):
            for Pkg in cls.PACKAGES_PATH:
                absPath = os.path.join(Pkg, Path)
                if os.path.exists(absPath):
                    return Pkg
        return Ws

    @classmethod
    def handleWsMacro(cls: 'MultipleWorkspace', PathStr: str) -> str:
        """Handle the $(WORKSPACE) tag.

        If current workspace is an invalid path relative to the tool, replace it.

        Args:
            cls (MultipleWorkspace): The class pointer
            PathStr (str): The path string

        Returns:
            (Str): Path string including the $(WORKSPACE)
        """
        warnings.warn("MultipleWorkspace is deprecated. Manually replace the $(WORKSPACE). If you believe "
                      "this functionality needs a direct replacement, file an issue in edk2-pytool-extensions.",
                      DeprecationWarning, stacklevel=2)
        TAB_WORKSPACE = '$(WORKSPACE)'
        if TAB_WORKSPACE in PathStr:
            PathList = PathStr.split()
            if PathList:
                for i, str in enumerate(PathList):
                    MacroStartPos = str.find(TAB_WORKSPACE)
                    if MacroStartPos != -1:
                        Substr = str[MacroStartPos:]
                        Path = Substr.replace(TAB_WORKSPACE, cls.WORKSPACE).strip()
                        if not os.path.exists(Path):
                            for Pkg in cls.PACKAGES_PATH:
                                Path = Substr.replace(TAB_WORKSPACE, Pkg).strip()
                                if os.path.exists(Path):
                                    break
                        PathList[i] = str[0:MacroStartPos] + Path
            PathStr = ' '.join(PathList)
        return PathStr

    @classmethod
    def getPkgPath(cls: 'MultipleWorkspace') -> list[str]:
        """Get all package paths.

        Args:
            cls (MultipleWorkspace): class pointer

        Returns:
            (list[str]): Packages Path

        """
        return cls.PACKAGES_PATH



# @file repo_resolver.py
# This module supports git operations (git repos).
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This module supports all git operations for edk2-pytool-extensions.

edk2-pytool-extensions has consolidated all git functionality within the repo_resolver module, which relies
on the GitPython PyPi module.  This module provides functionality to clone, checkout, and clean repos and
submodules along with providing common information about the repo or submodule.

The intent is to keep all git functionality consolidated in this module. Currently edk2_ci_setup.py,
edk2_setup.py, and git_dependency.py use this module to perform git operations.
"""
import logging
import os
from pathlib import Path
from typing import Optional

from edk2toollib.utility_functions import version_compare
from git import GitCommandError, InvalidGitRepositoryError, NoSuchPathError, Repo
from git.cmd import Git
from git.util import rmtree

from edk2toolext import edk2_logging

logger = logging.getLogger(__name__)
MIN_GIT_VERSION = "2.11.0"


def resolve(
    file_system_path: os.PathLike,
    dependency: dict,
    force: bool=False,
    ignore: bool=False,
    update_ok: bool=False
) -> None:
    """Resolves a particular repo.

    Args:
        file_system_path (os.Pathlike): path to repo
        dependency (dict): contains Path, Commit, Branch
        force (bool): If it is OK to update the commit or branch
        ignore (bool): If it is OK to ignore errors or not.
        update_ok (bool): If it is OK to update the commit or branch

    Raises:
        (Exception): An error resolving a repo and ignore=False
    """
    logger.info("Checking for dependency {0}".format(dependency["Path"]))
    git_path = os.path.abspath(file_system_path)

    # check if we have a path in our dependency
    if "Path" in dependency and not git_path.endswith(os.path.relpath(dependency["Path"])):
        # if we don't already the the path from the dependency at the end of the path we've been giving
        git_path = os.path.join(git_path, dependency["Path"])
    logger.info(f"Resolving at {git_path}")

    ##########################################################################
    #         Determine if we can, and how to checkout the repo              #
    ##########################################################################

    details = repo_details(git_path)

    ##########################################################################
    # 1. The directory does not exist, we can clone into it                  #
    ##########################################################################
    if not details["Path"].is_dir():
        logger.info(f"Cloning at {git_path}")
        clone_repo(git_path, dependency)
        checkout(git_path, dependency, True, False)
        return

    ##########################################################################
    # 2. The directory is empty, we can clone and checkout.                  #
    ##########################################################################
    if not any(details["Path"].iterdir()):
        clone_repo(git_path, dependency)
        checkout(git_path, dependency, True, False)
        return

    ##########################################################################
    # 3. Files are present, but it is not a git repo. Only clone and         #
    #    checkout if force is True.                                          #
    ##########################################################################
    if not details["Initialized"]:
        if force:
            clear_folder(git_path)
            logger.warning(
                f"Folder {git_path} is not a git repo and is being overwritten!")
            clone_repo(git_path, dependency)
            checkout(git_path, dependency, True, False)
            return
        else:
            if (ignore):
                logger.warning(
                    f"Folder {git_path} is not a git repo but Force parameter not used.  "
                    "Ignore State Allowed.")
                return
            else:
                logger.critical(
                    f"Folder {git_path} is not a git repo and it is not empty.")
                raise Exception(
                    f"Folder {git_path} is not a git repo and it is not empty")

    ##########################################################################
    # 4. A git repo exists, but it is dirty. Only re-clone and checkout if   #
    #    force is True.                                                      #
    ##########################################################################
    if details["Dirty"]:
        if force:
            clear_folder(git_path)
            logger.warning(
                f"Folder {git_path} is a git repo but is dirty and is being overwritten as requested!")
            clone_repo(git_path, dependency)
            checkout(git_path, dependency, True, False)
            return
        else:
            if (ignore):
                logger.warning(
                    f"Folder {git_path} is a git repo but is dirty and Force parameter not used.  "
                    "Ignore State Allowed.")
                return
            else:
                logger.critical(
                    f"Folder {git_path} is a git repo and is dirty.")
                raise Exception(
                    f"Folder {git_path} is a git repo and is dirty.")

    ##########################################################################
    # 5. The origin of the repo does not match. Only re-clone from the       #
    #    correct origin and checkout if force is True.                       #
    ##########################################################################
    if details["Url"] != dependency["Url"]:
        if force:
            clear_folder(git_path)
            logger.warning(
                f"Folder {git_path} is a git repo but it is at a different repo and is "
                "being overwritten as requested!")
            clone_repo(git_path, dependency)
            checkout(git_path, dependency, True, False)
            return
        else:
            if ignore:
                logger.warning(
                    f"Folder {git_path} is a git repo pointed at a different remote.  "
                    "Can't checkout or sync state")
                return
            else:
                logger.critical("The URL of the git Repo {2} in the folder {0} does not match {1}".format(
                    git_path, dependency["Url"], details["Url"]))
                raise Exception("The URL of the git Repo {2} in the folder {0} does not match {1}".format(
                    git_path, dependency["Url"], details["Url"]))

    ##########################################################################
    # 6. The repo is normal, Perform a regular checkout.                     #
    ##########################################################################
    checkout(git_path, dependency, update_ok, ignore, force)
    return


def resolve_all(
    workspace_path: os.PathLike,
    dependencies: list[dict],
    force: bool=False,
    ignore: bool=False,
    update_ok: bool=False,
    omnicache_dir: str=None
) -> list[str]:
    """Resolves all repos.

    Args:
        workspace_path (osPathlike): workspace root
        dependencies (list[dict]): Dict contains Path, Commit, Branch
        force (bool): If it is OK to update the commit or branch
        ignore (bool): If it is OK to ignore errors or not.
        update_ok (bool): If it is OK to update the commit or branch
        omnicache_dir (str): Omnicache path, if used

    Returns:
        (list[str]): list of paths to repos resolved

    Raises:
        (Exception): An error resolving a repo and ignore=False
    """
    repos = []
    if force:
        logger.info("Resolving dependencies by force")
    if update_ok:
        logger.info("Resolving dependencies with updates as needed")
    for dependency in dependencies:
        dep_path = dependency["Path"]
        logger.log(edk2_logging.PROGRESS, f"Syncing {dep_path}")
        if "ReferencePath" not in dependency and omnicache_dir:
            dependency["ReferencePath"] = omnicache_dir
        if "ReferencePath" in dependency:  # make sure that the omnicache dir is relative to the working directory
            dependency["ReferencePath"] = os.path.join(workspace_path, dependency["ReferencePath"])
        git_path = os.path.join(workspace_path, dep_path)
        repos.append(git_path)
        resolve(git_path, dependency, force, ignore, update_ok)

    # print out the details- this is optional
    for dependency in dependencies:
        git_path = os.path.join(workspace_path, dependency["Path"])
        details = repo_details(git_path)
        # print out details
        logger.info("{3} = Git Details: Url: {0} Branch {1} Commit {2}".format(
            details["Url"], details["Branch"], details["Head"]["HexSha"], dependency["Path"]))

    return repos


def repo_details(abs_file_system_path: os.PathLike) -> dict:
    """Return information about the repo.

    if self.valid is False, all other return members are set to default values
    and should not be expected to be correct.

    Args:
        abs_file_system_path (os.PathLike): repo directory

    Returns:
        (dict): dict containing details about the repository
    """
    git_version = ".".join(map(str, Git().version_info))
    if version_compare(git_version, MIN_GIT_VERSION) < 0:
        e = f"Upgrade Git! Current version is {git_version}. Minimum is {MIN_GIT_VERSION}"
        logging.error(e)
        raise RuntimeError(e)

    details = {
        "Path": Path(abs_file_system_path),
        "Valid": False,
        "GitVersion": git_version,
        "Initialized": False,
        "Bare": False,
        "Dirty": False,
        "Url": None,
        "Head": None,
        "Branch": None,
        "Submodules": [],
        "Remotes": [],
        "Worktrees": []
    }

    try:
        with Repo(abs_file_system_path) as repo:
            # Active Branch
            details["Branch"] = 'HEAD' if repo.head.is_detached else repo.active_branch.name

            # Worktrees
            worktree_list = []
            worktrees = repo.git.worktree("list", "--porcelain")
            for worktree in filter(lambda worktree: worktree.startswith("worktree"), worktrees.split('\n')):
                worktree_list.append(Path(worktree.split(" ")[1]))
                details["Worktrees"] = worktree_list

            # head information
            details["Head"] = {"HexSha": str(repo.head.commit.hexsha), "HexShaShort": str(repo.head.commit.hexsha[:7])}
            details["Valid"] = True
            details["Bare"] = repo.bare
            details["Dirty"] = repo.is_dirty(untracked_files=True)
            details["Initialized"] = True
            details["Submodules"] = [submodule.name for submodule in repo.submodules]
            details["Remotes"] = [remote.name for remote in repo.remotes]

            # 1. Use the remote associated with the branch if on a branch and it exists
            if not repo.head.is_detached and repo.branches[repo.head.ref.name].tracking_branch():
                remote_name = repo.branches[repo.head.ref.name].tracking_branch().remote_name
                details["Url"] = repo.remotes[remote_name].url
            # 2. Use the origin url if it exists
            elif "origin" in repo.remotes:
                details["Url"] = repo.remotes.origin.url
            # 3. Use whatever the first remote is
            elif len(repo.remotes) > 0:
                details["Url"] = repo.remotes[0].url

    except (InvalidGitRepositoryError, NoSuchPathError):
        pass
    return details


def clear_folder(abs_file_system_path: os.PathLike) -> None:
    """Cleans the folder.

    Args:
        abs_file_system_path (os.PathLike): Directory to delete.
    """
    logger.warning("WARNING: Deleting contents of folder {0} to make way for Git repo".format(
        abs_file_system_path))
    rmtree(abs_file_system_path)


def clone_repo(abs_file_system_path: os.PathLike, DepObj: dict) -> tuple:
    """Clones the repo in the folder using the dependency object.

    Args:
        abs_file_system_path (os.PathLike): destination to clone
        DepObj (dict): dict containing Commit, Full, Branch, etc

    Returns:
        ((PathLike, Bool)): (destination, result)
    """
    logger.info("Cloning repo: {0}".format(DepObj["Url"]))
    dest = abs_file_system_path

    if not os.path.isdir(dest):
        os.makedirs(dest, exist_ok=True)

    # Generate the clone flags
    shallow = False
    branch = None
    reference = None

    if "Commit" in DepObj:
        shallow = False
    if "Full" in DepObj and DepObj["Full"] is True:
        shallow = False
    if "Branch" in DepObj:
        shallow = True
        branch = DepObj["Branch"]
    if "ReferencePath" in DepObj and os.path.exists(DepObj["ReferencePath"]):
        reference = Path(DepObj["ReferencePath"])

    # Used to generate clone params from flags
    def _build_params_list(branch: str=None, shallow: str=None, reference: str=None) -> None:
        params = []
        if branch:
            shallow = True
            params.append('--branch')
            params.append(branch)
            params.append('--single-branch')
        if shallow:
            params.append('--depth=5')
        if reference:
            params.append('--reference')
            params.append(reference.as_posix())
        else:
            params.append("--recurse-submodules")  # if we don't have a reference we can just recurse the submodules
        return params

    # Run the command
    try:
        repo = Repo.clone_from(DepObj["Url"], dest, multi_options=_build_params_list(branch, shallow, reference))
    except GitCommandError:
        repo = None

    if repo is None:
        if "ReferencePath" not in DepObj:
            return (dest, False)

        # attempt a retry without the reference
        logger.warning("Reattempting to clone without a reference. {0}".format(DepObj["Url"]))

        try:
            repo = Repo.clone_from(DepObj["Url"], dest, multi_options=_build_params_list(branch, shallow))
        except GitCommandError:
            return (dest, False)

    # Repo cloned, perform submodule update if necessary
    if reference:
        repo.git.submodule('update', '--init', '--recursive', '--reference', reference)
    repo.close()
    return (dest, True)


def checkout(
    abs_file_system_path: str,
    dep: dict,
    update_ok: bool=False,
    ignore_dep_state_mismatch: bool=False,
    force: bool=False
) -> None:
    """Checks out a commit or branch.

    Args:
        abs_file_system_path (os.PathLike): The path to the repo
        dep (dict): A dictionary containing a either a Commit or Branch, and also a Path
        update_ok (bool): If it is OK to update the commit or branch
        ignore_dep_state_mismatch (bool): Whether a mismatch will result in an exception or not.
        force (bool): If it is OK to update the commit or branch

    Raises:
        (Exception): dependency state mismatch if ignore_dep_state_mismatch = False
        (GitCommandError): If the commit or branch does not exist locally and on the remote
    """
    details = repo_details(abs_file_system_path)
    with Repo(abs_file_system_path) as repo:
        if "Commit" in dep:
            commit = dep["Commit"]
            if update_ok or force:
                repo.git.fetch()
                try:
                    repo.git.checkout(commit)
                except GitCommandError:
                    repo.git.fetch()
                    repo.git.checkout(commit)
                repo.git.submodule("update", "--init", "--recursive")
            else:
                head = details["Head"]
                if commit in [head["HexSha"], head["HexShaShort"]]:
                    logger.debug(
                        f"Dependency {dep['Path']} state ok without update")
                elif ignore_dep_state_mismatch:
                    logger.warning(
                        f"Dependency {dep['Path']} is not in sync with requested commit.  Ignore state allowed")
                else:
                    logger.critical(
                        f"Dependency {dep['Path']} is not in sync with requested commit.  Fail.")
                    raise Exception(
                        f"Dependency {dep['Path']} is not in sync with requested commit.  Fail.")
            return

        elif "Branch" in dep:
            branch = dep["Branch"]
            if update_ok or force:
                repo.git.fetch()
                try:
                    repo.git.checkout(branch)
                except GitCommandError:
                    # try to fetch it and try to checkout again
                    logger.info("We failed to checkout this branch, we'll try to fetch")
                    # since the repo may have been cloned with --single-branch, add a refspec for the target branch.
                    refspec = "refs/heads/{0}:refs/remotes/origin/{0}".format(branch)
                    repo.remotes.origin.config_writer.set_value("fetch", refspec).release()
                    repo.git.fetch()
                    repo.git.checkout(branch)
                repo.git.submodule("update", "--init", "--recursive")
            else:
                if details["Branch"] == dep["Branch"]:
                    logger.debug(
                        f"Dependency {dep['Path']} state ok without update")
                elif ignore_dep_state_mismatch:
                    logger.warning(
                        f"Dependency {dep['Path']} is not in sync with requested branch.  Ignore state allowed")
                else:
                    error = "Dependency {0} is not in sync with requested branch. Expected: {1}. Got {2} Fail.".format(
                        dep["Path"], dep["Branch"], details["Branch"])
                    logger.critical(error)
                    raise Exception(error)
            return

        else:
            raise Exception("Branch or Commit must be specified for {0}".format(dep["Path"]))


def clean(abs_file_system_path: os.PathLike, ignore_files: Optional[list]=[]) -> None:
    """Resets and cleans the repo.

    Args:
        abs_file_system_path (os.PathLike): repo directory
        ignore_files (list): list of files to ignore when performing a clean

    Raises:
        (GitCommandError): The command is invalid
        (InvalidGitRepositoryError): The repo is in an invalid format
        (NoSuchPathError): The path does not exist
    """
    with Repo(abs_file_system_path) as repo:
        repo.git.reset("--hard")
        params = ["-xffd"]
        for file in ignore_files:
            params.append("-e")
            params.append(file)
        repo.git.clean(*params)


def submodule_clean(abs_file_system_path: os.PathLike, submodule: dict) -> None:
    """Resets and cleans a submodule of the repo.

    Args:
        abs_file_system_path (os.PathLike): repo directory
        submodule (dict): object containing path (relative) attribute

    Raises:
        (GitCommandError): The command is invalid
        (InvalidGitRepositoryError): The repo is in an invalid format
        (NoSuchPathError): The path does not exist
        (ValueError): submodule's path was invalid
    """
    with Repo(abs_file_system_path) as repo:
        submodule_match = next(filter(lambda s: Path(s.path) == Path(submodule.path), repo.submodules), None)

        if submodule_match is None:
            raise ValueError(f"Submodule {submodule.path} does not exist")

        if submodule_match.module_exists():
            clean(os.path.join(abs_file_system_path, submodule.path))


def submodule_resolve(
    abs_file_system_path: os.PathLike,
    submodule: dict,
    omnicache_path: Optional[os.PathLike]=None
) -> None:
    """Resolves a submodule to the specified branch and commit in .gitmodules.

    On the submodule, first performs a submodule sync followed by a submodule update --init.

    Args:
        abs_file_system_path (os.PathLike): repo directory
        submodule (dict): object containing attributes: path (relative) and recursive
        omnicache_path (PathLike | None): absolute path to the omnicache, if used

    Raises:
        (GitCommandError): The command is invalid
        (InvalidGitRepositoryError): The repo is in an invalid format
        (NoSuchPathError): The path does not exist
    """
    with Repo(abs_file_system_path) as repo:

        logger.debug(f'Syncing {submodule.path}')
        repo.git.submodule('sync', '--', submodule.path)

        params = ['update', '--init']
        if submodule.recursive:
            params.append("--recursive")
        if omnicache_path:
            params.append('--reference')
            params.append(omnicache_path)
        params.append(submodule.path)
        logger.debug(f'Updating {submodule.path}')
        repo.git.submodule(*params)

    with Repo(Path(abs_file_system_path, submodule.path)) as _:
        logger.debug(f'{submodule.path} is valid and resolved.')



# @file edk2_logging.py
# Handle basic logging config for invocables;
# splits logs into a master log and per package.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Handles basic logging config for invocables.

edk2_logging will automatically filter logs for PATs / Secrets when it is
detected that the invocable is running on a CI system. It does this via
searching for "CI" or "TF_BUILD" in the os's environment variables. If either
of these exists and and are set to TRUE, filtering will occur.

Splits logs into a master log and per package log.
"""
import logging
import os
import re
import shutil
from typing import Optional, TextIO, Union

try:
    from edk2toollib.log import ansi_handler
except ImportError:
    ansi_handler = None
try:
    from edk2toollib.log import string_handler
except ImportError:
    string_handler = None
try:
    from edk2toollib.log import file_handler
except ImportError:
    file_handler = logging


# These three are for emitting different events
# section is for marking different sections of the build process
# subsection is similar to sub section but denotes a subsection of the current section
# both of the section levels are high enough that they won't get filtered out
# progress is for marking things like a process completed. Similar to critical but doesn't mean the process is exiting
# progress is below critical so it can be turned off but still high enough that it doesn't get filtered out
SECTION = logging.CRITICAL + 2  # just above critical
SUB_SECTION = logging.CRITICAL + 1  # just above critical
PROGRESS = logging.CRITICAL - 1  # just below critical


# sub_directory is relative to ws argument
def clean_build_logs(ws: str, sub_directory: Optional[str]=None) -> None:
    """Removes all build logs."""
    # Make sure that we have a clean environment.
    if sub_directory is None:
        sub_directory = os.path.join("Build", "BuildLogs")
    if os.path.isdir(os.path.join(ws, sub_directory)):
        shutil.rmtree(os.path.join(ws, sub_directory))


def get_section_level() -> int:
    """Returns SECTION."""
    return SECTION


def get_subsection_level() -> int:
    """Returns SUB_SECTION."""
    return SUB_SECTION


def get_progress_level() -> int:
    """Returns PROGRESS."""
    return PROGRESS


def get_edk2_filter(verbose: bool=False) -> logging.Filter:
    """Returns an edk2 filter."""
    gEdk2Filter = Edk2LogFilter()
    if verbose:
        gEdk2Filter.setVerbose(verbose)
    return gEdk2Filter


def log_progress(message: str) -> None:
    """Creates a logging message at the progress section level."""
    logging.log(get_progress_level(), message)


def setup_section_level() -> None:
    """Sets up different sections to log to."""
    # todo define section level
    # add section as a level to the logger
    section_level = get_section_level()
    subsection_level = get_subsection_level()
    progress_level = get_progress_level()
    if logging.getLevelName(section_level) != "SECTION":
        logging.addLevelName(section_level, "SECTION")
    if logging.getLevelName(subsection_level) != "SUBSECTION":
        logging.addLevelName(subsection_level, "SUBSECTION")
    if logging.getLevelName(progress_level) != "PROGRESS":
        logging.addLevelName(progress_level, "PROGRESS")


# creates the the plaintext logger
def setup_txt_logger(
    directory: str,
    filename: str="log",
    logging_level: int=logging.INFO,
    formatter: Optional[logging.Formatter]=None,
    logging_namespace: Optional[str]='',
    isVerbose: bool=False
) -> tuple:
    """Configures a text logger."""
    logger = logging.getLogger(logging_namespace)
    log_formatter = formatter
    if log_formatter is None:
        log_formatter = logging.Formatter("%(levelname)s - %(message)s")

    if not os.path.isdir(directory):
        os.makedirs(directory)

    logfile_path = os.path.join(directory, filename + ".txt")

    # delete file before starting a new log
    if os.path.isfile(logfile_path):
        os.remove(logfile_path)

    # Create file logger
    filelogger = file_handler.FileHandler(filename=(logfile_path), mode='a')
    filelogger.setLevel(logging_level)
    filelogger.setFormatter(log_formatter)
    logger.addHandler(filelogger)

    filelogger.addFilter(get_edk2_filter(isVerbose))

    return logfile_path, filelogger


# sets up a colored console logger
def setup_console_logging(
    logging_level: int=logging.INFO,
    formatter: Optional[logging.Formatter]=None,
    logging_namespace: Optional[str]='',
    isVerbose: bool=False,
    use_azure_colors: bool=False,
    use_color: bool=True
) -> logging.Handler:
    """Configures a console logger.

    Filtering of secrets will automatically occur if "CI" or "TF_BUILD" is set to TRUE
    in the os's environment.
    """
    if formatter is None and isVerbose:
        formatter_msg = "%(name)s: %(levelname)s - %(message)s"
    elif formatter is None:
        formatter_msg = "%(levelname)s - %(message)s"
    else:
        formatter_msg = formatter

    formatter = logging.Formatter(formatter_msg)

    # create a safe handler so that any logging emitted when creating the ansi logger is handled
    safeHandler = logging.StreamHandler()
    safeHandler.setLevel(logging_level)
    safeHandler.addFilter(get_edk2_filter(isVerbose))
    safeHandler.setFormatter(formatter)
    logger = logging.getLogger(logging_namespace)
    logger.addHandler(safeHandler)

    # create the ansi logger if needed
    if use_azure_colors or use_color and ansi_handler:
        formatter = ansi_handler.ColoredFormatter(formatter_msg, use_azure=use_azure_colors)
        coloredHandler = ansi_handler.ColoredStreamHandler()
        coloredHandler.setLevel(logging_level)
        coloredHandler.addFilter(get_edk2_filter(isVerbose))
        coloredHandler.setFormatter(formatter)
        # make sure to remove the safe handler so we don't have two handlers
        logger.removeHandler(safeHandler)
        logger.addHandler(coloredHandler)
        return coloredHandler
    # return the safe handler if we didn't create a colored handler
    return safeHandler


def stop_logging(loghandle: Union[list[logging.Handler], logging.Handler], logging_namespace: Optional[str]='') -> None:
    """Stops logging on a log handle."""
    logger = logging.getLogger(logging_namespace)
    if loghandle is None:
        return
    if isinstance(loghandle, list):
        # if it's an array, process each element as a handle
        for handle in loghandle:
            handle.close()
            logger.removeHandler(handle)
    else:
        loghandle.close()
        logger.removeHandler(loghandle)


def create_output_stream(level: int=logging.INFO, logging_namespace: Optional[str]='') -> logging.Handler:
    """Creates an output stream to log to."""
    # creates an output stream that is in memory
    if string_handler:
        handler = string_handler.StringStreamHandler()
    else:
        handler = logging.StreamHandler()
    logger = logging.getLogger(logging_namespace)
    handler.setLevel(level)
    logger.addHandler(handler)
    return handler


def remove_output_stream(handler: logging.Handler, logging_namespace: Optional[str]='') -> None:
    """Removes an output stream to log to."""
    logger = logging.getLogger(logging_namespace)
    if isinstance(handler, list):
        for single_handler in handler:
            logger.removeHandler(single_handler)
    else:
        logger.removeHandler(handler)


def scan_compiler_output(output_stream: TextIO) -> list[tuple]:
    """Scans the compiler for errors and warnings.

    Returns:
        (list[tuple[logging.Type, str]]): list of tuples containing the type
            of issue (Error, warning) and the description.
    """
    # seek to the start of the output stream
    def output_compiler_error(match: re.Match, line: str, start_txt:str="Compiler") ->str:
        start, end = match.span()
        source = line[:start].strip()
        error = line[end:].strip()
        num = match.group(1)
        return f"{start_txt} #{num} from {source} {error}"
    problems = []
    output_stream.seek(0, 0)
    error_exp = re.compile(r"error [A-EG-Z]?(\d+):")
    # Would prefer to do something like r"(?:\/[^\/: ]*)+\/?:\d+:\d+: (error):"
    # but the script is currently setup on fixed formatting assumptions rather
    # than offering per rule flexibility to parse tokens via regex
    gcc_error_exp = re.compile(r":\d+:\d+: (error):", re.IGNORECASE)
    gcc_fatal_error_exp = re.compile(r"fatal error:")
    edk2_error_exp = re.compile(r"error F(\d+):")
    build_py_error_exp = re.compile(r"error (\d+)E:")
    linker_error_exp = re.compile(r"error LNK(\d+):")
    warning_exp = re.compile(r"warning [A-Z]?(\d+):")
    rust_error_exp = re.compile(r"^(-->|error(?:\[[A-Za-z]\d+\])?:)")
    for raw_line in output_stream.readlines():
        line = raw_line.strip("\n").strip()
        match = error_exp.search(line)
        if match is not None:
            error = output_compiler_error(match, line, "Compiler")
            problems.append((logging.ERROR, error))
        match = gcc_error_exp.search(line)
        if match is not None:
            error = output_compiler_error(match, line, "Compiler")
            problems.append((logging.ERROR, error))
        match = gcc_fatal_error_exp.search(line)
        if match is not None:
            problems.append((logging.ERROR, line))
        match = warning_exp.search(line)
        if match is not None:
            error = output_compiler_error(match, line, "Compiler")
            problems.append((logging.WARNING, error))
        match = linker_error_exp.search(line)
        if match is not None:
            error = output_compiler_error(match, line, "Linker")
            problems.append((logging.ERROR, error))
        match = edk2_error_exp.search(line)
        if match is not None:
            error = output_compiler_error(match, line, "EDK2")
            problems.append((logging.ERROR, error))
        match = build_py_error_exp.search(line)
        if match is not None:
            error = output_compiler_error(match, line, "Build.py")
            problems.append((logging.ERROR, error))
        match = rust_error_exp.search(line)
        if match is not None:
            problems.append((logging.ERROR, line))
    return problems


class Edk2LogFilter(logging.Filter):
    """Subclass of logging.Filter."""
    _allowedLoggers = ["root", "git.cmd", "edk2toolext.environment.repo_resolver"]

    def __init__(self) -> None:
        """Inits a filter."""
        logging.Filter.__init__(self)
        self._verbose = False
        self._currentSection = "root"
        self.apply_filter = False

        # Turn on filtering for azure pipelines / github actions
        if os.environ.get("CI", "FALSE").upper() == "TRUE" or os.environ.get("TF_BUILD", "FALSE").upper() == "TRUE":
            self.apply_filter = True

        secrets_regex_strings = [
            r"[a-z0-9]{46}",  # Nuget API Key is generated as all lowercase
            r"gh[pousr]_[A-Za-z0-9_]+",  # Github PAT
            r"[a-z0-9]{52}",  # Azure PAT is generated as all lowercase
        ]

        self.secrets_regex = re.compile(r"{}".format("|".join(secrets_regex_strings)), re.IGNORECASE)

    def setVerbose(self, isVerbose: bool=True) -> None:
        """Sets the filter verbosity."""
        self._verbose = isVerbose

    def addSection(self, section: str) -> None:
        """Adds a section to the filter."""
        # TODO request the global singleton?
        # how to make this class static
        Edk2LogFilter._allowedLoggers.append(section)

    def filter(self, record: logging.LogRecord) -> bool:
        """Adds a filter for a record if it doesn't already exist."""
        # check to make sure we haven't already filtered this record
        if record.name not in Edk2LogFilter._allowedLoggers and record.levelno < logging.WARNING and not self._verbose:
            return False
        if self.apply_filter:
            record.msg = self.secrets_regex.sub("*******", str(record.msg))
        return True



# @file edk2_invocable
# Class Object providing Edk2 specific Invocable functionality
#
# This should not be instantiated but should be used as baseclass
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

"""Edk2 Invocable Interface to be overriden in a subclass.

Provides two main classes, the Edk2InvocableSettingsInterface and the
Edk2Invocable, which should be used as subclasses to create invocables that
perform tasks associated with the EDK2 build system. Any Edk2Invocable subclass
should be platform agnostic and work for any platform. Platform specific data
is provided via the Edk2InvocableSettingsInterface.
"""
import argparse
import inspect
import logging
import os
import sys
import warnings
from random import choice
from string import ascii_letters
from textwrap import dedent
from typing import Iterable

import pkg_resources
from edk2toollib.utility_functions import GetHostInfo, RunCmd, import_module_by_file_name, locate_class_in_module

from edk2toolext.base_abstract_invocable import BaseAbstractInvocable
from edk2toolext.environment import shell_environment, version_aggregator


class Edk2InvocableSettingsInterface():
    """Settings APIs to support an Edk2Invocable.

    This is an interface definition only to show which functions are
    required to be implemented and can be implemented in a settings
    manager.

    !!! example "Example of Overriding Edk2InvocableSettingsInterface"
        ``` python
        import os
        import logging
        import argparse
        from typing import Iterable, Tuple
        from edk2toolext.edk2_invocable import Edk2InvocableSettingsInterface

        class NewInvocableSettingsManager(Edk2InvocableSettingsInterface):
            def GetWorkspaceRoot(self) -> str:
                return os.path.abspath(__file__)
            def GetPackagesPath(self) -> Iterable[os.PathLike]
                return ["C:/src/MU_BASECORE", "Common/MU"]
            def GetActiveScopes(self) -> Tuple[str]:
                return ("edk2-build", "pipbuild-win")
            def GetLoggingLevel(self, loggerType: str) -> str:
                if loggerType == 'txt':
                    return logging.WARNING
                else:
                    return logging.INFO
            def AddCommandLineOptions(self, parserObj: object) -> None:
                parserObj.add_argument('-a', "--arch", dest="build_arch", type=str, default="IA32,X64")
            def RetrieveCommandLineOptions(self, args: object) -> None:
                self.arch = args.build_arch
            def GetSkippedDirectories(self) -> Tuple[str]:
                return ("Downloads/Extra")
        ```

    !!! warning
        This interface should not be subclassed directly unless you are creating a new invocable
        Other invocables subclass from this interface, so you have the ability to
        call the functions in this class as a part of those invocable settings managers.
    """

    def GetWorkspaceRoot(self) -> str:
        """Return the workspace root for initializing the SDE, will be mapped as edk2 WORKSPACE.

        !!! tip
            Required Override in a subclass

        The absolute path to the root of the workspace

        Returns:
            (str): path to workspace root
        """
        raise NotImplementedError()

    def GetPackagesPath(self) -> Iterable[os.PathLike]:
        """Provides an iterable of paths which will be mapped as edk2 PACKAGES_PATH.

        !!! tip
            Optional Override in a subclass

        Path can be,
            1. Absolute or
            2. Relative to WORKSPACE or
            3. Relative to current working dir

        See Also,
            https://github.com/tianocore/tianocore.github.io/wiki/Multiple_Workspace

        Returns:
            (Iterable[os.PathLike]): paths
        """
        return []

    def GetActiveScopes(self) -> tuple[str]:
        """Provides scopes that should be active for this process for SDE.

        !!! tip
            Optional Override in a subclass

        Returns:
            (Tuple[str]): scopes
        """
        return ()

    def GetLoggingLevel(self, loggerType: str) -> int:
        """Get the logging level depending on logger type.

        !!! tip
            Optional Override in a subclass

        Returns:
            (Logging.Level): The logging level

        !!! note "loggerType possible values"
            "base": lowest logging level supported

            "con": logs to screen

            "txt": logs to plain text file
        """
        return None

    def AddCommandLineOptions(self, parserObj: argparse.ArgumentParser) -> None:
        """Add command line options to the argparser.

        !!! tip
            Optional override in a subclass

        Args:
            parserObj (argparse.ArgumentParser): Argparser object.
        """

    def RetrieveCommandLineOptions(self, args: argparse.Namespace) -> None:
        """Retrieve Command line options from the argparser.

        !!! tip
            Optional override in a subclass

        Args:
            args (argparse.Namespace): argparser args namespace containing command line options
        """

    def GetSkippedDirectories(self) -> tuple[str]:
        """Returns a tuple containing workspace-relative directories to be skipped by SDE.

        !!! tip
            Optional override in a subclass

        Returns:
            (Tuple[str]): directories to be skipped.
        """
        return ()


class Edk2Invocable(BaseAbstractInvocable):
    """Base class for Edk2 based invocables.

    Edk2 means it has common features like workspace, packagespath,
    scopes, and other name value pairs

    Attributes:
        PlatformSettings (Edk2InvocableSettingsInterface): A settings class
        PlatformModule (object): The platform module
        Verbose (bool): CLI Argument to determine whether or not to have verbose

    !!! tip
        Checkout BaseAbstractInvocable Attributes to find any additional attributes that might exist.

    !!! warning
        This Invocable should only be subclassed if creating a new invocable
    """

    @classmethod
    def collect_python_pip_info(cls: 'Edk2Invocable') -> None:
        """Class method to collect all pip packages names and versions.

        Reports them to the global version_aggregator as well as print them to the screen.
        """
        # Get the current python version
        cur_py = "%d.%d.%d" % sys.version_info[:3]
        ver_agg = version_aggregator.GetVersionAggregator()
        ver_agg.ReportVersion("Python", cur_py, version_aggregator.VersionTypes.TOOL)
        # Get a list of all the packages currently installed in pip
        pip_packages = [p for p in pkg_resources.working_set]
        # go through all installed pip versions
        for package in pip_packages:
            version = pkg_resources.get_distribution(package).version
            logging.info("{0} version: {1}".format(package.project_name, version))
            ver_agg.ReportVersion(package.project_name, version, version_aggregator.VersionTypes.PIP)

    @classmethod
    def collect_rust_info(cls: 'Edk2Invocable') -> None:
        """Class method to collect Rust tool versions.

        Reports them to the global version_aggregator as well as print them to the screen.
        """
        import re
        from io import StringIO

        def get_rust_tool_version(tool_name: str, tool_params: str = "--version") -> str:
            cmd_output = StringIO()
            ret = RunCmd(tool_name, tool_params, outstream=cmd_output, logging_level=logging.DEBUG)
            if ret == 0:
                return cmd_output.getvalue().strip()
            else:
                return "N/A"

        tools = {
            "cargo": ("cargo",),
            "cargo make": ("cargo", "make --version"),
            "rustc": ("rustc",)
        }

        for tool_name, tool_cmd in tools.items():
            ver = get_rust_tool_version(*tool_cmd)
            match = re.search(r'(\d+\.\d+\.\d+)', ver)
            if match:
                ver = match.group(1)
            elif ver != "N/A":
                raise Exception("A Rust tool is installed, but its version "
                                "format is unexpected and cannot be parsed.")

            logging.info(f"{tool_name} version: {ver}")
            ver_agg = version_aggregator.GetVersionAggregator()
            ver_agg.ReportVersion(tool_name, ver, version_aggregator.VersionTypes.TOOL)

    def GetWorkspaceRoot(self) -> os.PathLike:
        """Returns the absolute path to the workspace root.

        !!! note
            Workspace Root is platform specific and thus provided by the PlatformSettings

        Returns:
            absolute path to workspace root
        """
        try:
            return self.PlatformSettings.GetWorkspaceRoot()
        except AttributeError:
            raise RuntimeError("Can't call this before PlatformSettings has been set up!")

    def GetPackagesPath(self) -> Iterable[os.PathLike]:
        """Returns an iterable of packages path.

        !!! note
            PackagesPath is platform specific and thus provided by the PlatformSettings

        Returns:
            Packages path
        """
        try:
            return [x for x in self.PlatformSettings.GetPackagesPath() if x not in self.GetSkippedDirectories()]
        except AttributeError:
            raise RuntimeError("Can't call this before PlatformSettings has been set up!")

    def GetActiveScopes(self) -> tuple[str]:
        """Returns an iterable of Active scopes.

        !!! note
            Scopes are platform specific and thus provided by the PlatformSettings

        This function adds an os specific scope in addition to scopes provided by SettingsManager

        Returns:
            active scopes
        """
        try:
            scopes = self.PlatformSettings.GetActiveScopes()
        except AttributeError:
            raise RuntimeError("Can't call this before PlatformSettings has been set up!")

        # Add any OS-specific scope.
        if GetHostInfo().os == "Windows":
            scopes += ('global-win',)
        elif GetHostInfo().os == "Linux":
            scopes += ('global-nix',)
        # Add the global scope. To be deprecated.
        scopes += ('global',)
        return scopes

    def GetLoggingLevel(self, loggerType: str) -> int:
        """Get the logging level for a given logger type.

        !!! note
            Logging Level is platform specific and thus provided by the PlatformSettings

        Returns:
            (logging.Level): logging level
        """
        if self.Verbose:
            return logging.DEBUG

        try:
            level = self.PlatformSettings.GetLoggingLevel(loggerType)
            if level is not None:
                return level
        except Exception:
            pass

        if loggerType == "con":
            return logging.WARNING
        return logging.DEBUG

    def AddCommandLineOptions(self, parserObj: argparse.ArgumentParser) -> None:
        """Add command line options to the argparser.

        !!! note
            Optional Override to add functionality
        """

    def RetrieveCommandLineOptions(self, args: argparse.Namespace) -> None:
        """Retrieve command line options from the argparser.

        !!! note
            Optional Override to add functionality
        """

    def GetSkippedDirectories(self) -> tuple[str]:
        """Returns a Tuple containing workspace-relative directories that should be skipped.

        !!! tip
            Override in a subclass to add invocable specific directories to skip

        !!! note
            Skipped Directories are platform specific and thus provided by the PlatformSettings

        Returns:
            (Tuple[str]): skipped directories as relative paths
        """
        try:
            return self.PlatformSettings.GetSkippedDirectories()
        except AttributeError:
            raise RuntimeError("Can't call this before PlatformSettings has been set up!")

    def GetSettingsClass(self) -> type:
        """The required settings manager for the invocable.

        !!! note
            Required override to define Edk2InvocableSettingsInterface subclass specific to the invocable

        Returns:
            (Edk2InvocableSettingsInterface): Subclass of Edk2InvocableSettingsInterface
        """
        raise NotImplementedError()

    def GetLoggingFolderRelativeToRoot(self) -> str:
        """Directory containing all logging files."""
        return "Build"

    def AddParserEpilog(self) -> str:
        """Adds an epilog to the end of the argument parser when displaying help information.

        Returns:
            (str): The string to be added to the end of the argument parser.
        """
        epilog = dedent('''\
            CLI Env Guide:
              <key>=<value>              - Set an env variable for the pre/post build process
              <key>                      - Set a non-valued env variable for the pre/post build process
              BLD_*_<key>=<value>        - Set a build flag for all build types
                                           (key=value will get passed to build process)
              BLD_*_<key>                - Set a non-valued build flag for all build types
              BLD_<TARGET>_<key>=<value> - Set a build flag for build type of <target>
                                           (key=value will get passed to build process for given build type)
              BLD_<TARGET>_<key>         - Set a non-valued build flag for a build type of <target>
            ''')
        return epilog

    def ParseCommandLineOptions(self) -> None:
        """Parses command line options.

        Sets up argparser specifically to get PlatformSettingsManager instance.
        Then sets up second argparser and passes it to child class and to PlatformSettingsManager.
        Finally, parses all known args and then reads the unknown args in to build vars.
        """
        # first argparser will only get settings manager and help will be disabled
        settingsParserObj = argparse.ArgumentParser(add_help=False)

        settingsParserObj.add_argument('-h', '--help', dest="help", action="store_true",
                                       help='show this help message and exit')
        settingsParserObj.add_argument('-c', '--platform_module', dest='platform_module',
                                       default="PlatformBuild.py", type=str,
                                       help='Provide the Platform Module relative to the current working directory.'
                                            f'This should contain a {self.GetSettingsClass().__name__} instance.')

        # get the settings manager from the provided file and load an instance
        settingsArg, unknown_args = settingsParserObj.parse_known_args()
        try:
            self.PlatformModule = import_module_by_file_name(os.path.abspath(settingsArg.platform_module))
            self.PlatformSettings = locate_class_in_module(
                self.PlatformModule, self.GetSettingsClass())()
        except (TypeError):
            # Gracefully exit if the file we loaded isn't the right type
            class_name = self.GetSettingsClass().__name__
            print(f"Unable to use {settingsArg.platform_module} as a {class_name}")
            print("Did you mean to use a different kind of invocable?")
            try:
                # If this works, we can provide help for whatever special functions
                # the subclass is offering.
                self.AddCommandLineOptions(settingsParserObj)
                Module = self.PlatformModule
                module_contents = dir(Module)
                # Filter through the Module, we're only looking for classes.
                classList = [getattr(Module, obj) for obj in module_contents if inspect.isclass(getattr(Module, obj))]
                # Get the name of all the classes
                classNameList = [obj.__name__ for obj in classList]
                # TODO improve filter to no longer catch imports as well as declared classes
                imported_classes = ", ".join(classNameList)  # Join the classes together
                print(f"The module you imported contains {imported_classes}")
            except Exception:
                raise

            settingsParserObj.print_help()
            sys.exit(1)

        except (FileNotFoundError):
            if settingsArg.help:
                try:
                    print("WARNING: Some command line arguments may be missing. Provide a PLATFORM_MODULE file to "
                          "ensure all command line arguments are present.\n")
                    self.AddCommandLineOptions(settingsParserObj)
                except Exception:
                    pass
            else:
                # Gracefully exit if we can't find the file
                print(f"We weren't able to find {settingsArg.platform_module}")
            settingsParserObj.print_help()
            sys.exit(2)

        except Exception as e:
            print(f"Error: We had trouble loading {settingsArg.platform_module}. Is the path correct?")
            # Gracefully exit if setup doesn't go well.
            settingsParserObj.print_help()
            print(e)
            sys.exit(2)

        # Turn on Deprecation warnings for code in the module
        warnings.filterwarnings("default", category=DeprecationWarning, module=self.PlatformModule.__name__)

        # instantiate the second argparser that will get passed around
        parserObj = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,)

        # first pass it to the subclass
        self.AddCommandLineOptions(parserObj)

        # next pass it to the settings manager
        self.PlatformSettings.AddCommandLineOptions(parserObj)

        default_build_config_path = os.path.join(self.GetWorkspaceRoot(), "BuildConfig.conf")

        # add the common stuff that everyone will need
        parserObj.add_argument('--build-config', dest='build_config', default=default_build_config_path, type=str,
                               help='Provide shell variables in a file')
        parserObj.add_argument('--verbose', '--VERBOSE', '-v', dest="verbose", action='store_true', default=False,
                               help='Overrides platform settings and sets all loggers to verbose (logging.DEBUG).')

        # set the epilog to display with --help, -h
        parserObj.epilog = self.AddParserEpilog()

        # setup sys.argv and argparse round 2
        sys.argv = [sys.argv[0]] + (["--help"] if settingsArg.help else unknown_args)
        args, unknown_args = parserObj.parse_known_args()
        self.Verbose = args.verbose

        # give the parsed args to the subclass
        self.RetrieveCommandLineOptions(args)

        # give the parsed args to platform settings manager
        self.PlatformSettings.RetrieveCommandLineOptions(args)

        #
        # Look through unknown_args and BuildConfig for strings that are:
        # 1. x=y, -> set env.SetValue(x, y),
        # 2. x, -> set env.SetValue(x, random_string)
        # then remove these items from the list.
        #
        # Non valued build variables (#2) set the value to a random string
        # as the expectation is that any developer using this functionality
        # check for the existence of the build variable rather then the value
        # of the variable. This is to have parity between edk2's build -D
        # flag and stuart.
        env = shell_environment.GetBuildVars()
        BuildConfig = os.path.abspath(args.build_config)

        for argument in unknown_args:
            if argument.count("=") == 1:
                tokens = argument.strip().split("=")
                env.SetValue(tokens[0].strip().upper(), tokens[1].strip(), "From CmdLine")
            elif argument.count("=") == 0 and not argument.startswith(("-", "/")):
                env.SetValue(argument.strip().upper(),
                             ''.join(choice(ascii_letters) for _ in range(20)),
                             "Non valued variable set From cmdLine")
            else:
                print(f"error: unexpected argument: [{argument}]. Pass --help for command information.")
                sys.exit(-1)

        unknown_args.clear()  # remove the arguments we've already consumed

        if os.path.isfile(BuildConfig):
            with open(BuildConfig) as file:
                for line in file:
                    stripped_line = line.strip().partition("#")[0]
                    if len(stripped_line) == 0:
                        continue
                    unknown_args.append(stripped_line)

        for argument in unknown_args:
            if argument.count("=") == 1:
                tokens = argument.strip().split("=")
                env.SetValue(tokens[0].strip().upper(), tokens[1].strip(), "From BuildConf")
            elif argument.count("=") == 0:
                env.SetValue(argument.strip().upper(),
                             ''.join(choice(ascii_letters) for _ in range(20)),
                             "Non valued variable set from BuildConfig")
            else:
                raise RuntimeError(f"Unknown variable passed in via BuildConfig: {argument}")



# @file edk2_parse.py
# An invocable to run workspace parsers on a workspace and generate a database.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An invocable to run workspace parsers on a workspace and generate a database."""
import argparse
import logging
import os
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Iterable, Optional, Sequence

import yaml
from edk2toollib.database import Edk2DB
from edk2toollib.database.tables import (
    EnvironmentTable,
    InfTable,
    InstancedFvTable,
    InstancedInfTable,
    PackageTable,
    SourceTable,
    TableGenerator,
)
from edk2toollib.uefi.edk2.path_utilities import Edk2Path
from edk2toollib.utility_functions import import_module_by_file_name, locate_class_in_module

from edk2toolext.environment import shell_environment
from edk2toolext.environment.uefi_build import UefiBuilder
from edk2toolext.environment.var_dict import VarDict
from edk2toolext.invocables.edk2_multipkg_aware_invocable import (
    Edk2MultiPkgAwareInvocable,
    MultiPkgAwareSettingsInterface,
)

DB_NAME = "DATABASE.db"


class AppendSplitAction(argparse.Action):
    """An argparse action to split a comma separated list and append it to a list.

    example: -p Pkg1,Pkg2 -p Pkg3 => ['Pkg1', 'Pkg2', 'Pkg3']
    """
    def __call__(
            self,
            parser: ArgumentParser,
            namespace: Namespace,
            values: Sequence[str],
            option_string: Optional[str] = None
        ) -> None:
        """The command to invoke the action."""
        items = getattr(namespace, self.dest, [])
        items.extend(values.split(','))
        setattr(namespace, self.dest, items)


class ParseSettingsManager(MultiPkgAwareSettingsInterface):
    """Settings to support ReportSettingsManager functionality."""
    def GetPackagesSupported(self) -> Iterable[str]:
        """Returns an iterable of edk2 packages supported by this build."""
        # Re-define and return an empty list instead of raising an exception as this is only needed when parsing
        # based off of a CI Settings File.
        return []
    def GetArchitecturesSupported(self) -> Iterable:
        """Returns an iterable of edk2 architectures supported by this build."""
        # Re-define and return an empty list instead of raising an exception as this is only needed when parsing
        # based off of a CI Settings File.
        return []
    def GetTargetsSupported(self) -> Iterable:
        """Returns an iterable of edk2 target tags supported by this build."""
        # Re-define and return an empty list instead of raising an exception as this is only needed when parsing
        # based off of a CI Settings File.
        return []


class Edk2Parse(Edk2MultiPkgAwareInvocable):
    """An invocable used to parse the environment and generate a database.

    This invocable has the ability to parse both a repo containing generic Edk2 packages, and a repo containing
    platform packages. When parsing generic packages, The Edk2Report invocable uses methods from the
    MultiPkgAwareSettingsInterface, most commonly found ina CI Settings File. In this scenario, it will either parse
    all packages defined by `GetPackagesSupported()` or a subset of those packages as defined by the -p flag. The same
    can be said about the architectures by using `GetDefinedArchitectures()` and the -a command.

    When the goal is to parse a platform package, the Edk2Report invocable uses methods from the `UefiBuilder`, which
    is most commonly found in the Platform Build File. In this scenario, it will invoke the methods found in the
    `UefiBuilder` which are used to build a platform. It will not execute any plugins, build the platform, nor execute
    any functionality normally called after the build command.

    Addionally, any build variables can be overwritten or added via the command line VAR=VALUE functionality.
    """

    def __init__(self) -> None:
        """Initializes an Edk2Report Object."""
        super().__init__()
        self.is_uefi_builder = False

    def GetSettingsClass(self) -> type:
        """Returns the Settings Manager for the invocable."""
        return ParseSettingsManager

    def GetVerifyCheckRequired(self) -> bool:
        """Will call self_describing_environment.VerifyEnvironment if this returns True."""
        return False

    def AddCommandLineOptions(self, parserObj: ArgumentParser) -> None:
        """Adds the command line options."""
        super().AddCommandLineOptions(parserObj) # Adds the CI Settings File options
        parserObj.add_argument('--clear', '--Clear', "--CLEAR", dest='clear', action='store_true',
                               help="Deletes the database before parsing the environment.")
        parserObj.add_argument('--source-stats', '-S', dest='source_stats', action='store_true', default=False,
                               help="Uses pygount to generate code statistics for each parsed source file, including "
                               "lines of code, comments, and blank lines. Greatly increases parse time.")
        parserObj.add_argument('-l', '--load-table', '--Load-Table', '--LOAD-TABLE', dest='extra_tables',
                               action=AppendSplitAction, default=[], metavar='<path>',
                               help="Comma separated path to a python file containing a `TableParser`(s). "
                               "Includes this table with the default tables. Can be provided multiple times")

    def RetrieveCommandLineOptions(self, args: Namespace) -> None:
        """Retrives the command line options."""
        super().RetrieveCommandLineOptions(args) # Stores the CI Settings File options
        self.clear = args.clear
        self.is_uefi_builder = locate_class_in_module(self.PlatformModule, UefiBuilder) is not None
        self.extra_tables = args.extra_tables
        self.source_stats = args.source_stats
        self.args = args

    def GetLoggingFileName(self, loggerType: str) -> str:
        """Returns the logging file name for this invocation."""
        return "PARSE_LOG"

    def Go(self) -> int:
        """Executes the invocable. Runs the subcommand specified by the user."""
        logging.warning("stuart_parse is in active development. Please report any issues to the edk2-pytool-extensions "
                        "repo.")
        db_path = Path(self.GetWorkspaceRoot()) / self.GetLoggingFolderRelativeToRoot() / DB_NAME
        pathobj = Edk2Path(self.GetWorkspaceRoot(), self.GetPackagesPath())
        env = shell_environment.GetBuildVars()

        TABLES = [
            EnvironmentTable(),
            PackageTable(),
            SourceTable(source_stats=self.source_stats),
            InfTable(),
            InstancedInfTable(),
            InstancedFvTable(),
        ]

        if self.clear:
            db_path.unlink(missing_ok=True)

        db = Edk2DB(db_path, pathobj=pathobj)
        db.register(*TABLES)

        # Load and register any extra requested tables
        db.register(*self.load_extra_tables(self.extra_tables))

        # Generate environment aware tables
        if self.is_uefi_builder:
            self.parse_with_builder_settings(db, pathobj, env)
        else:
            self.parse_with_ci_settings(db, pathobj, env)

        logging.info(f'Database generated at {db_path}.')
        return 0

    def parse_with_builder_settings(self, db: Edk2DB, pathobj: Edk2Path, env: VarDict) -> int:
        """Parses the workspace using a uefi builder to setup the environment."""
        logging.info("Setting up the environment with the UefiBuilder.")
        exception_msg = ""
        try:
            platform_module = locate_class_in_module(self.PlatformModule, UefiBuilder)
            build_settings = platform_module()
            build_settings.RetrieveCommandLineOptions(self.args)
            build_settings.Clean = False
            build_settings.SkipPreBuild = True
            build_settings.SkipBuild = True
            build_settings.SkipPostBuild = True
            build_settings.FlashImage = False
            build_settings.Go(self.GetWorkspaceRoot(), os.pathsep.join(self.GetPackagesPath()),
                                self.helper, self.plugin_manager)
            build_settings.PlatformPreBuild()
        except Exception as e:
            exception_msg = e
        if exception_msg:
            logging.error("Failed to run UefiBuilder to set the environment.")
            logging.error(exception_msg)
            return -1

        env_dict = env.GetAllBuildKeyValues() | env.GetAllNonBuildKeyValues()
         # Log the environment for debug purposes
        for key, value in env_dict.items():
            logging.debug(f"  {key} = {value}")
        logging.info("Running parsers with the following settings:")
        logging.info(f"  TARGET: {env_dict['TARGET']}")
        logging.info(f"  ACTIVE_PLATFORM: {env_dict['ACTIVE_PLATFORM']}")
        logging.info(f"  TARGET_ARCH: {env_dict['TARGET_ARCH']}")
        db.parse(env_dict)
        return 0

    def parse_with_ci_settings(self, db: Edk2DB, pathobj: Edk2Path, env: VarDict) -> int:
        """Parses the workspace using ci settings to setup the environment."""
        for package in self.requested_package_list:
            for target in set(self.requested_target_list) & set(["DEBUG", "RELEASE", "NOOPT"]):
                logging.info(f"Setting up the environment for {package}.")
                shell_environment.CheckpointBuildVars()

                pkg_config = self._get_package_config(pathobj, package)
                dsc = pkg_config.get("CompilerPlugin", {"DscPath": ""})["DscPath"]
                if not dsc:
                    logging.info("Package skipped! No DSC found in ci.yaml file.")
                    continue
                dsc = str(Path(package, dsc))

                env.SetValue("TARGET", target, "Set via commandline arguments")
                env.SetValue("ACTIVE_PLATFORM", dsc, "Set Automatically.")
                env.SetValue("TARGET_ARCH", " ".join(self.requested_architecture_list), "Set Automatically")
                env.SetValue("PACKAGES_PATH", ";".join(self.GetPackagesPath()), "Set Automatically")

                # Load the Defines
                for key, value in pkg_config.get("Defines", {}).items():
                    env.SetValue(key, value, "Defined in Package CI yaml")

                # Log the environment for debug purposes
                for key, value in (env.GetAllBuildKeyValues() | env.GetAllNonBuildKeyValues()).items():
                    logging.debug(f"  {key} = {value}")

                env_dict = env.GetAllBuildKeyValues() | env.GetAllNonBuildKeyValues()
                logging.info("Running parsers with the following settings:")
                logging.info(f"  TARGET: {target}")
                logging.info(f"  ACTIVE_PLATFORM: {dsc}")
                logging.info(f"  TARGET_ARCH: {' '.join(self.requested_architecture_list)}")
                db.parse(env_dict)
                shell_environment.RevertBuildVars()
        return 0

    def _get_package_config(self, pathobj: Edk2Path, pkg: str) -> str:
        """Gets configuration information for a package from the ci.yaml file."""
        pkg_config_file = pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(
            str(Path(pkg, pkg + ".ci.yaml"))
        )
        if pkg_config_file:
            with open(pkg_config_file, 'r') as f:
                return yaml.safe_load(f)
        else:
            logging.debug(f"No package config file for {pkg}")

    def load_extra_tables(self, table_paths: list) -> list[TableGenerator]:
        """Loads TableParsers from a python file.

        Returns:
            (list): A list of instantiated TableParsers.
        """
        table_list = []
        for path in table_paths:
            if not Path(path).exists():
                logging.warning(f'[{path}] does not exist; Skipping.')
                continue
            try:
                module = import_module_by_file_name(path)
                table_generator = locate_class_in_module(module, TableGenerator)()
                table_list.append(table_generator)
            except TypeError:
                logging.warning(f'[{path}] does not contain a TableGenerator class; Skipping.')
            except SyntaxError as e:
                logging.warning(f'Failed to register [{path}] due to a SyntaxError; Error:')
                logging.warning(f'  {e}')
        return table_list

def main() -> None:
    """Entry point to invoke Edk2PlatformSetup."""
    Edk2Parse().Invoke()



# @file image_validation.py
# This tool allows a user validate an PE/COFF file
# against specific requirements
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This tool allows the user to validate a PE/COFF file against specific requirements.

It also provides CLI functions to set and clear the nx_compat flag.
"""

import argparse
import logging
import os
import sys
from typing import Optional, Sequence

from pefile import MACHINE_TYPE, PE, SECTION_CHARACTERISTICS, SUBSYSTEM_TYPE

from edk2toolext import edk2_logging

########################
#    Helper Functions  #
########################


def has_characteristic(data: int, mask: int) -> bool:
    """Checks if data has a specific mask."""
    return ((data & mask) == mask)


def set_bit(data: int, bit: int) -> int:
    """Sets a specific bit."""
    return data | (1 << bit)


def clear_bit(data: int, bit: int) -> int:
    """Clears a specific bit."""
    return data & ~(1 << bit)


def set_nx_compat_flag(pe: PE) -> PE:
    """Sets the nx_compat flag to 1 in the PE/COFF file."""
    dllchar = pe.OPTIONAL_HEADER.DllCharacteristics
    dllchar = set_bit(dllchar, 8)  # 8th bit is the nx_compat_flag
    pe.OPTIONAL_HEADER.DllCharacteristics = dllchar
    pe.merge_modified_section_data()
    return pe


def get_nx_compat_flag(pe: PE) -> PE:
    """Reads the nx_compat flag of the PE/COFF file."""
    dllchar = pe.OPTIONAL_HEADER.DllCharacteristics

    if has_characteristic(dllchar, 256):  # 256 (8th bit) is the mask
        logging.info('True')
        return 1
    else:
        logging.info('False')
        return 0


def clear_nx_compat_flag(pe: PE) -> PE:
    """Sets the nx_compat flag to 0 in the PE/COFF file."""
    dllchar = pe.OPTIONAL_HEADER.DllCharacteristics
    dllchar = clear_bit(dllchar, 8)  # 8th bit is the nx_compat_flag
    pe.OPTIONAL_HEADER.DllCharacteristics = dllchar
    pe.merge_modified_section_data()
    return pe


def fill_missing_requirements(default: dict, target: dict) -> dict:
    """Fills missing requirements for a specific test config with default config.

    As an example, If there are specific requirements for an APP PE/COFF, those
    will take override the default requirements.Any requirement not specified
    by the APP config will be filled by the DEFAULT config.
    """
    for key in default:
        if key not in target:
            target[key] = default[key]
    return target


class Result:
    """Test results."""
    PASS = '[PASS]'
    WARN = '[WARNING]'
    SKIP = '[SKIP]'
    FAIL = '[FAIL]'


class TestInterface:
    """Interface for creating tests to execute on parsed PE/COFF files."""
    def name(self) -> str:
        """Returns the name of the test.

        "WARNING: Implement in a subclass.
        """
        raise NotImplementedError("Must Override Test Interface")

    def execute(self, pe: PE, config_data: dict) -> Result:
        """Executes the test on the pefile.

        Arguments:
            pe (PE): a parsed PE/COFF image file
            config_data (dict): config data for the test

        Returns:
            (Result): SKIP, WARN, FAIL, PASS

        WARNING: Implement in a subclass.
        """
        raise NotImplementedError("Must Override Test Interface")


class TestManager(object):
    """Manager responsible for executing all tests on all parsed PE/COFF files."""
    def __init__(self, config_data: Optional[dict]=None) -> None:
        """Inits the TestManager with configuration data.

        Args:
            config_data (dict, optional): the configuration data, loads default
                data if not provided.
        """
        self.tests = []
        if config_data:
            self.config_data = config_data
        else:
            self.config_data = {
                "TARGET_ARCH": {
                    "X64": "IMAGE_FILE_MACHINE_AMD64",
                    "IA32": "IMAGE_FILE_MACHINE_I386",
                    "AARCH64": "IMAGE_FILE_MACHINE_ARM64",
                    "ARM": "IMAGE_FILE_MACHINE_ARM"
                },
                "IMAGE_FILE_MACHINE_AMD64": {
                    "DEFAULT": {
                        "DATA_CODE_SEPARATION": True,
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_ROM"
                        ],
                        "ALIGNMENT": [
                            {
                                "COMPARISON": "==",
                                "VALUE": 4096
                            }
                        ]
                    },
                    "APP": {
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_APPLICATION"
                        ]
                    },
                    "DRIVER": {
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER"
                            "IMAGE_SUBSYSTEM_EFI_ROM"
                        ]
                    },
                },
                "IMAGE_FILE_MACHINE_ARM": {
                    "DEFAULT": {
                        "DATA_CODE_SEPARATION": True,
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_ROM"
                        ],
                        "ALIGNMENT": [
                            {
                                "COMPARISON": "==",
                                "VALUE": 4096
                            }
                        ]
                    },
                    "APP": {
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_APPLICATION"
                        ],
                    },
                    "DRIVER": {
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_ROM"
                        ]}
                },
                "IMAGE_FILE_MACHINE_ARM64": {
                    "DEFAULT": {
                        "DATA_CODE_SEPARATION": True,
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_ROM"
                        ],
                        "ALIGNMENT": [
                            {
                                "COMPARISON": "==",
                                "VALUE": 4096
                            }
                        ]
                    },
                    "APP": {
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_APPLICATION"
                        ]
                    },
                    "DRIVER": {
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_ROM"
                        ]
                    }
                },
                "IMAGE_FILE_MACHINE_I386": {
                    "DEFAULT": {
                        "DATA_CODE_SEPARATION": True,
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_ROM"
                        ],
                        "ALIGNMENT": [
                            {
                                "COMPARISON": "==",
                                "VALUE": 4096
                            }
                        ]
                    },
                    "APP": {
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_APPLICATION"
                        ]
                    },
                    "DRIVER": {
                        "ALLOWED_SUBSYSTEMS": [
                            "IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER",
                            "IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER"
                            "IMAGE_SUBSYSTEM_EFI_ROM"
                        ]
                    }
                }
            }

    def add_test(self, test: TestInterface) -> None:
        """Adds a test to the test manager.

        Will be executed in the order added.

        Args:
            test (TestInterface): A subclasses of the TestInterface
        """
        self.tests.append(test)

    def add_tests(self, tests: list[TestInterface]) -> None:
        """Adds multiple test to the test manager.

        Tests will be executed in the order added.

        Args:
            tests (List[TestInterface]): A list of subclasses of the TestInterface
        """
        self.tests.extend(tests)

    def run_tests(self, pe: PE, profile: str="DEFAULT") -> Result:
        """Runs all tests that have been added to the test manager.

        Tests will be executed in the order added

        Args:
            pe (PE): The parsed pe
            profile (str):  profile to lookup in the config data

        Returns:
            (Result.PASS): All tests passed successfully (including warnings)
            (Result.SKIP): There is no information in the config file for the target and fv file type
            (Result.FAIL): At least one test failed. Error messages can be found in the log
        """
        # Catch any invalid profiles
        machine_type = MACHINE_TYPE[pe.FILE_HEADER.Machine]
        if not self.config_data[machine_type].get(profile):
            logging.error(f'Profile type {profile} is invalid. Exiting...')
            return Result.FAIL

        # Fill any missing configurations for the specific module type with the default
        default = self.config_data[machine_type]["DEFAULT"]
        target = self.config_data[machine_type][profile]
        target_requirements = fill_missing_requirements(default, target)

        target_info = {
            "MACHINE_TYPE": machine_type,
            "PROFILE": profile
        }
        test_config_data = {
            "TARGET_INFO": target_info,
            "TARGET_REQUIREMENTS": target_requirements
        }

        logging.debug(f'Executing tests with settings [{machine_type}][{profile}]')
        overall_result = Result.PASS
        for test in self.tests:
            logging.debug(f'Starting test: [{test.name()}]')

            result = test.execute(pe, test_config_data)

            # Overall Result can only go lower (Pass -> Warn -> Fail)
            if result == Result.PASS:
                logging.debug(f'{result}')
            elif result == Result.SKIP:
                logging.debug(f'{result}: No Requirements for [{machine_type}][{profile}]')
            elif overall_result == Result.PASS:
                overall_result = result
            elif overall_result == Result.WARN and result == Result.FAIL:
                overall_result = result

        return overall_result


###########################
#       TESTS START       #
###########################
class TestWriteExecuteFlags(TestInterface):
    """Section data / code separation verification Test.

    This test ensures that each section of the binary is not both
    write-able and execute-able. Sections can only be one or the other
    (or neither).This test is done by iterating over each section and
    checking the characteristics label for the Write Mask (0x80000000)
    and Execute Mask (0x20000000).

    Output:
        @Success: Only one (or neither) of the two masks (Write, Execute) are
            present
        @Skip: Test Skipped per config
        @Fail: Both the Write and Execute flags are present
            Possible Solution:

    Possible Solution:
        Update the failed section's characteristics to ensure it is either
        Write-able or Read-able, but not both.
    """

    def name(self) -> str:
        """Returns the name of the test."""
        return 'Section data / code separation verification'

    def execute(self, pe: PE, config_data: dict) -> Result:
        """Executes the test on the pefile.

        Arguments:
            pe (PE): a parsed PE/COFF image file
            config_data (dict): config data for the test

        Returns:
            (Result): SKIP, WARN, FAIL, PASS
        """
        target_requirements = config_data["TARGET_REQUIREMENTS"]

        if target_requirements.get("DATA_CODE_SEPARATION", False) is False:
            return Result.SKIP

        for section in pe.sections:
            if (has_characteristic(section.Characteristics, SECTION_CHARACTERISTICS["IMAGE_SCN_MEM_EXECUTE"])
               and has_characteristic(section.Characteristics, SECTION_CHARACTERISTICS["IMAGE_SCN_MEM_WRITE"])):

                logging.error(f'[{Result.FAIL}]: Section [{section.Name.decode().strip()}] \
                              should not be both Write and Execute')
                return Result.FAIL
        return Result.PASS


class TestSectionAlignment(TestInterface):
    """Section alignment verification Test.

    Checks the section alignment of the binary by accessing the optional
    header, then the section alignment. This value must meet the
    requirements specified in the config file.

    Output:
        @Success: Image alignment meets the requirement specified in the
            config file
        @Warn: Image Alignment value is not found in the Optional Header or
            set to 0
        @Skip: No Alignment requirements specified in the config file
        @Fail: Image alignment does not meet the requirements specified in
            the config file

    Possible Solution:
        Update the section alignment of the binary to match the
        requirements specified in the config file
    """

    def name(self) -> str:
        """Returns the name of the test."""
        return 'Section alignment verification'

    def execute(self, pe: PE, config_data: dict) -> Result:
        """Executes the test on the pefile.

        Arguments:
            pe (PE): a parsed PE/COFF image file
            config_data (dict): config data for the test

        Returns:
            (Result): SKIP, WARN, FAIL, PASS
        """
        target_requirements = config_data["TARGET_REQUIREMENTS"]
        target_info = config_data["TARGET_INFO"]

        alignments = target_requirements.get("ALIGNMENT")
        if alignments is None or len(alignments) == 0:
            return Result.SKIP

        try:
            alignment = pe.OPTIONAL_HEADER.SectionAlignment
        except Exception:
            logging.warning("Section Alignment is not present")
            return Result.WARN

        if alignment is None or alignment == 0:
            return Result.WARN

        if len(alignments) > 1:
            logical_separator = target_requirements.get("ALIGNMENT_LOGIC_SEP")

            if logical_separator is None:
                logging.error("Multiple alignment requirements exist, but no logical separator provided")
                return Result.FAIL
            elif logical_separator == "AND":
                result = True
                for reqs in alignments:
                    result = result and eval(f'{alignment} {reqs["COMPARISON"]} {reqs["VALUE"]}')
            elif logical_separator == "OR":
                result = False
                for reqs in alignments:
                    result = result or eval(f'{alignment} {reqs["COMPARISON"]} {reqs["VALUE"]}')
            else:
                logging.error("Invalid logical separator provided")
                return Result.FAIL
        else:
            req = alignments[0]
            result = eval(f'{alignment} {req["COMPARISON"]} {req["VALUE"]}')
        if result is False:
            logging.error(f'[{Result.FAIL}: Section Alignment Required: \
                            [{target_info["MACHINE_TYPE"]}] \
                            [{target_info["PROFILE"]}]: \
                            [(Detected): {alignment}]')
            return Result.FAIL

        return Result.PASS


class TestSubsystemValue(TestInterface):
    """Subsystem type verification Test.

    Checks the subsystem value by accessing the optional header, then
    subsystem value. This value must match one of the allowed subsystem
    described in the config file

    Output:
        @Success: Subsystem type found in the optional header matches one of
            the allowed subsystems
        @Warn   : Subsystem type is not found in the optional header
        @Skip   : No subsystem type restrictions specified
        @Fail   : Subsystem type found in the optional header does not match
            one of the allowed subsystems

    Possible Solution:
        Update the subsystem type in the source code.
    """

    def name(self) -> str:
        """Returns the name of the test."""
        return 'Subsystem type verification'

    def execute(self, pe: PE, config_data: dict) -> Result:
        """Executes the test on the pefile.

        Arguments:
            pe (PE): a parsed PE/COFF image file
            config_data (dict): config data for the test

        Returns:
            (Result): SKIP, WARN, FAIL, PASS
        """
        target_requirements = config_data["TARGET_REQUIREMENTS"]

        subsystems = target_requirements.get("ALLOWED_SUBSYSTEMS")
        if subsystems is None or len(subsystems) == 0:
            return Result.SKIP

        try:
            subsystem = pe.OPTIONAL_HEADER.Subsystem
        except Exception:
            logging.warn("Section Alignment is not present")
            return Result.WARN

        if subsystem is None:
            logging.warning(f'[{Result.WARN}]: Subsystem type is not present in the optional header.')
            return Result.WARN

        actual_subsystem = SUBSYSTEM_TYPE.get(subsystem)

        if actual_subsystem is None:
            logging.error(f'[{Result.WARN}]: Invalid Subsystem present')
            return Result.FAIL

        if actual_subsystem in subsystems:
            return Result.PASS
        else:
            logging.error(f'{Result.FAIL}: Submodule Type [{actual_subsystem}] not allowed.')
            return Result.FAIL
###########################
#        TESTS END        #
###########################


def get_cli_args(args: Sequence[str]) -> argparse.Namespace:
    """Adds CLI arguments for using the image validation tool."""
    parser = argparse.ArgumentParser(description='A Image validation tool for memory mitigation')

    parser.add_argument('-i', '--file',
                        type=str,
                        required=True,
                        help='path to the image that needs validated.')
    parser.add_argument('-d', '--debug',
                        action='store_true',
                        default=False)

    parser.add_argument('-p', '--profile',
                        type=str,
                        default=None,
                        help='the profile config to be verified against. \
                            Will use the default, if not provided')

    group = parser.add_mutually_exclusive_group()

    group.add_argument('--set-nx-compat',
                       action='store_true',
                       default=False,
                       help='sets the NX_COMPAT flag')

    group.add_argument('--clear-nx-compat',
                       action='store_true',
                       default=False,
                       help='clears the NX_COMPAT flag')

    group.add_argument('--get-nx-compat',
                       action='store_true',
                       default=False,
                       help='returns the value of the NX_COMPAT flag')

    return parser.parse_args(args)


def main() -> None:
    """Main entry point into the image validation tool."""
    # setup main console as logger
    logger = logging.getLogger('')
    logger.setLevel(logging.INFO)
    console = edk2_logging.setup_console_logging(False)
    logger.addHandler(console)

    args = get_cli_args(sys.argv[1:])

    if args.debug is True:
        console.setLevel(logging.DEBUG)

    # Set the nx compatability flag and exit
    if args.set_nx_compat is True:
        pe = PE(args.file)
        set_nx_compat_flag(pe)
        os.remove(args.file)
        pe.write(args.file)
        exit(0)

    # clear the nx compatability flag and exit
    if args.clear_nx_compat is True:
        pe = PE(args.file)
        clear_nx_compat_flag(pe)
        os.remove(args.file)
        pe.write(args.file)
        exit(0)

    # exit with status equal to if nx compatability is present or not
    if args.get_nx_compat is True:
        pe = PE(args.file)
        exit(get_nx_compat_flag(pe))

    test_manager = TestManager()
    test_manager.add_test(TestWriteExecuteFlags())
    test_manager.add_test(TestSectionAlignment())
    test_manager.add_test(TestSubsystemValue())

    pe = PE(args.file)
    if not args.profile:
        result = test_manager.run_tests(pe)
    else:
        result = test_manager.run_tests(pe, args.profile)

    logging.info(f'Overall Result: {result}')
    if result == Result.SKIP:
        logging.info('No Test requirements in the config file for this file.')
    elif result == Result.PASS or result == Result.WARN:
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == '__main__':
    main()



# @file az_cli_universal_dependency.py
# This module implements ExternalDependency for azure cli universal packages.
# More details here:
# https://docs.microsoft.com/en-us/azure/devops/artifacts/quickstarts/universal-packages?view=azure-devops
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An ExternalDependency subclass able to download from an Azure feed."""
import json
import logging
import os
import shutil
from io import StringIO
from pathlib import Path

from edk2toollib.utility_functions import RemoveTree, RunCmd

from edk2toolext.environment import shell_environment, version_aggregator
from edk2toolext.environment.extdeptypes.web_dependency import WebDependency
from edk2toolext.environment.external_dependency import ExternalDependency


class AzureCliUniversalDependency(ExternalDependency):
    """An ExternalDependency subclass able to download from an Azure feed.

    Attributes:
        feed (str): feed name
        version (str): semantic version <Major.Minor.Patch>
        source (str): url of organization (example: https://dev.azure.com/tianocore)
        project (str): <name of project for project scoped feed.  If missing assume organization scoped>
        name (str): name of artifact
        file-filter (str): <optional> filter for folders and files.
        compression_type (str): <optional> Compression type used, if compressed.
        internal_path (str): <optional> Path inside the compressed file, if the ext_dep is compressed
        pat_var (str): shell_var name for PAT for this ext_dep

    !!! tip
        The attributes are what must be described in the ext_dep yaml file!
    """
    TypeString = "az-universal"

    # https://docs.microsoft.com/en-us/azure/devops/cli/log-in-via-pat?view=azure-devops&tabs=windows
    AZURE_CLI_DEVOPS_ENV_VAR = "AZURE_DEVOPS_EXT_PAT"

    VersionLogged = False

    @classmethod
    def VerifyToolDependencies(cls: 'AzureCliUniversalDependency') -> None:
        """Verify any tool environment or dependencies requirements are met.

        Log to Version Aggregator the Tool Versions
        """
        if cls.VersionLogged:
            return
        results = StringIO()
        RunCmd("az", "--version", outstream=results, raise_exception_on_nonzero=True)
        results.seek(0)

        to_find = ["azure-cli", "azure-devops"]  # find these keys in the version output
        found = dict()

        for line in results.readlines():
            if len(to_find) == 0:
                break

            for f in to_find:
                if f in line:
                    found[f] = line.split()[1]
                    to_find.remove(f)
                    break

        # Log the versions found
        for (k, v) in found.items():
            version_aggregator.GetVersionAggregator().ReportVersion(k, v, version_aggregator.VersionTypes.TOOL)

        # Check requirements

        # 1 - az cli tool missing will raise exception on call to az --version earlier in function

        # 2 - Check for azure-devops extension
        if 'azure-devops' not in found.keys():
            logging.critical("Missing required Azure-cli extension azure-devops")
            raise EnvironmentError("Missing required Azure-cli extension azure-devops")

        cls.VersionLogged = True

    def __init__(self, descriptor: dict) -> None:
        """Inits a Azure CLI dependency based off the provided descriptior."""
        super().__init__(descriptor)
        self.global_cache_path = None
        self.organization = self.source
        self.feed = descriptor.get('feed')
        self.project = descriptor.get('project', None)
        self.file_filter = descriptor.get('file-filter', None)
        self.compression_type = descriptor.get('compression_type', None)
        self.internal_path = descriptor.get('internal_path', "/")
        if self.internal_path:
            self.internal_path = os.path.normpath(self.internal_path)
            self.internal_path = self.internal_path.strip(os.path.sep)
        _pat_var = descriptor.get('pat_var', None)
        self._pat = None

        if _pat_var is not None:
            # Get the PAT or if not defined in shell_var it will return None
            self._pat = shell_environment.GetEnvironment().get_shell_var(_pat_var)

    def _fetch_from_cache(self, package_name: str) -> bool:
        ## AZ tool has no cache feature
        return False

    def __str__(self) -> str:
        """Return a string representation."""
        return f"AzCliUniversalDependency: {self.name}@{self.version}"

    def _attempt_universal_install(self, install_dir: str) -> None:
        #
        # fetch the contents of the package.
        #
        cmd = ["az"]
        cmd += ["artifacts", "universal", "download"]
        cmd += ["--organization", self.organization]
        cmd += ["--feed", self.feed]
        cmd += ["--name", self.name]
        cmd += ["--version", self.version]
        if self.project is not None:
            cmd += ["--project", self.project]
            cmd += ["--scope", "project"]
        if self.file_filter is not None:
            cmd += ["--file-filter", self.file_filter]
        cmd += ["--path", '"' + install_dir + '"']

        # get the shell environment
        e = os.environ.copy()
        # if PAT then add the PAT as special variable
        if self._pat is not None:
            e[self.AZURE_CLI_DEVOPS_ENV_VAR] = self._pat

        results = StringIO()
        RunCmd(cmd[0], " ".join(cmd[1:]), outstream=results, environ=e, raise_exception_on_nonzero=True)
        # az tool returns json data that includes the downloaded version
        # lets check it to double confirm
        result_data = json.loads(results.getvalue())
        results.close()
        downloaded_version = result_data['Version']
        if self.version != downloaded_version:
            self.version = downloaded_version  # set it so state file is accurate and will fail on verify
            raise Exception("Download Universal Package version (%s) different than requested (%s)." %
                            (downloaded_version, self.version))

    def fetch(self) -> None:
        """Fetches the dependency using internal state from the init."""
        #
        # Before trying anything with we should check
        # to see whether the package is already in
        # our local cache. If it is, we avoid a lot of
        # time and network cost by copying it directly.
        #
        if self._fetch_from_cache(self.name):
            # We successfully found the package in the cache.
            # The published path may change now that the package has been unpacked.
            self.published_path = self.compute_published_path()
            return

        #
        # If we are still here, the package wasn't in the cache.
        # We need to ask Nuget to find it.
        #
        temp_directory = self.get_temp_dir()
        self._attempt_universal_install(temp_directory)

        #
        # if there is a compression type, the only thing downloaded must be the compressed
        # file. If this is true, we will unpack it directly into the the contents
        # directory and delete the temp directory.
        #
        if self.compression_type:
            files = list(Path(temp_directory).iterdir())
            if len(files) != 1:
                raise Exception("Expected only 1 file in the downloaded directory")
            tmp_file_path = files[0]
            WebDependency.unpack(tmp_file_path, temp_directory, self.internal_path, self.compression_type)
            tmp_file_path.unlink()

            source_dir = os.path.join(temp_directory, self.internal_path)
        else:
            source_dir = temp_directory
        #
        # Next, copy the contents of the package to the
        # final resting place.
        #
        shutil.copytree(source_dir, self.contents_dir)

        RemoveTree(temp_directory)

        #
        # Add a file to track the state of the dependency.
        #
        self.update_state_file()

        # The published path may change now that the package has been unpacked.
        self.published_path = self.compute_published_path()

    def get_temp_dir(self) -> str:
        """Returns the temporary directory the Azure CLI feed is downloaded to."""
        return self.contents_dir + "_temp"

    def clean(self) -> None:
        """Removes the temporary directory the NuGet package is downloaded to."""
        super(AzureCliUniversalDependency, self).clean()
        if os.path.isdir(self.get_temp_dir()):
            RemoveTree(self.get_temp_dir())



# @file edk2_multipkg_aware_invocable
# An intermediate class that supports a multi-package aware
# invocable process.
#
# Add cmdline parameter handling, a base settings manager class,
# and a Callback.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An intermediate class that supports a multi-package aware invocable process.

Provides two main classes, the MultiPkgAwareSettingsInterface and the
Edk2MultiPkgAwareInvocable that act as an intermediate class that other
invocables that require a multi-package aware invocable process. These classes
should only be subclassed if a new invocable is being developed. Any
Edk2MultiPkgAwareInvocable should be platform agnostic and work for any
platform. Platform specific data is provided via the
MultiPkgAwareSettingsInterface
"""
import argparse
from typing import Iterable

from edk2toolext.edk2_invocable import Edk2Invocable, Edk2InvocableSettingsInterface


class MultiPkgAwareSettingsInterface(Edk2InvocableSettingsInterface):
    """Settings to support Multi-Pkg functionality.

    This is an interface definition only to show which functions are required
    to be implemented and which functions can be implemented.

    !!! example " Example of Overriding MultiPkgAwareSettingsInterface"
        ``` python
        import os
        import logging
        import argparse
        from typing import Iterable, Tuple
        from edk2toolext.edk2_multipkg_aware_invocable import MultiPkgAwareSettingsInterface
        class NewInvocableSettingsManager(MultiPkgAwareSettingsInterface):

            def GetPackagesSupported(self):
                return ("PlatformPkg",)

            def GetArchitecturesSupported(self):
                return ("IA32","X64")

            def GetTargetsSupported(self):
                return ("TARGET", "RELEASE")

            def SetPackages(self, list_of_requested_packages):

                if len(filter(lambda pkg: pkg in self.GetPackagesSupported(), list_of_requested_packages)) !=
                   len(list_of_requested_packages):
                   raise Exception("Requested Packages contains unsupported Package")
                else:
                    self.pkgs = list_of_requested_packages

            def SetArchitectures(self, list_of_requested_architectures):
                if list_of_requested_architectures != self.GetPackagesSupported():
                    raise Exception("Only Support IA32,X64 combination")
            def SetTargets(self, list_of_requested_targets):
                if list_of_requested_targets != self.GetArchitecturesSupported():
                    raise Exception("Only Support "TARGET", "RELEASE combination")
        ```

    !!! warning
        This interface should not be subclassed directly unless creating a new invocable type. Override these
        methods as a part of other subclasses invocable settings managers such as SetupSettingsManager, etc.
    """

    # ####################################################################################### #
    #                           Supported Values and Defaults                                 #
    # ####################################################################################### #
    def GetPackagesSupported(self) -> Iterable[str]:
        """Returns an iterable of edk2 packages supported by this build.

        !!! tip
            Required Override in a subclass

        Returns:
            (Iterable): edk2 packages

        Note:
            packages should be relative to workspace or package path
        """
        raise NotImplementedError()

    def GetArchitecturesSupported(self) -> Iterable[str]:
        """Returns an iterable of edk2 architectures supported by this build.

        !!! tip
            Required Override in a subclass

        Returns:
            (Iterable): architectures (X64, I32, etc.)
        """
        raise NotImplementedError()

    def GetTargetsSupported(self) -> Iterable[str]:
        """Returns an iterable of edk2 target tags supported by this build.

        !!! tip
            Required Override in a subclass

        Returns:
            (Iterable): targets (DEBUG, RELEASE, etc)
        """
        raise NotImplementedError()

    # ####################################################################################### #
    #                     Verify and Save requested Config                                    #
    # ####################################################################################### #
    def SetPackages(self, list_of_requested_packages: list) -> None:
        """Confirms the requested package list is valid.

        !!! tip
            Optional Override in a subclass

        Args:
            list_of_requested_packages (list[str]): packages to be built

        Raises:
            Exception: A requested package is not supported
        """

    def SetArchitectures(self, list_of_requested_architectures: list) -> None:
        """Confirms the requested architecture list is valid.

        !!! tip
            Optional Override in a subclass

        Args:
            list_of_requested_architectures (list[str]): architectures to be built

        Raises:
            Exception: A requested architecture is not supported
        """

    def SetTargets(self, list_of_requested_target: list) -> None:
        """Confirms the requested target list is valid.

        !!! tip
            Optional Override in a subclass

        Args:
            list_of_requested_target (list[str]): targets to use

        Raises:
            Exception: A requested target is not supported
        """


class Edk2MultiPkgAwareInvocable(Edk2Invocable):
    """Base class for Multi-Pkg aware invocable.

    Attributes:
        requested_architecture_list (list): requested architectures to build
        requested_package_list (list): requested packages to build
        requested_target_list (list): requested targets to use

    !!! tip
        Checkout Edk2Invocable Attributes to find any additional attributes that might exist.

    !!! warning
        This invocable should only be subclassed if creating a new invocable
    """

    def __init__(self) -> None:
        """Initializes the Invocable."""
        self.requested_architecture_list = []
        self.requested_package_list = []
        self.requested_target_list = []
        super().__init__()

    def AddCommandLineOptions(self, parserObj: argparse.ArgumentParser) -> None:
        """Adds command line options to the argparser."""
        # This will parse the packages that we are going to update
        parserObj.add_argument('-p', '--pkg', '--pkg-dir', dest='packageList', type=str,
                               help='Optional - A package or folder you want to update (workspace relative).'
                               'Can list multiple by doing -p <pkg1>,<pkg2> or -p <pkg3> -p <pkg4>',
                               action="append", default=[])
        parserObj.add_argument('-a', '--arch', dest="requested_arch", type=str, default=None,
                               help="Optional - CSV of architecture requested to update. Example: -a X64,AARCH64")
        parserObj.add_argument('-t', '--target', dest='requested_target', type=str, default=None,
                               help="Optional - CSV of targets requested to update.  Example: -t DEBUG,NOOPT")

    def RetrieveCommandLineOptions(self, args: argparse.Namespace) -> None:
        """Retrieve command line options from the argparser ."""
        packageListSet = set()
        for item in args.packageList:  # Parse out the individual packages
            item_list = item.split(",")
            for individual_item in item_list:
                # in case cmd line caller used Windows folder slashes
                individual_item = individual_item.replace("\\", "/").rstrip("/")
                packageListSet.add(individual_item.strip())
        self.requested_package_list = list(packageListSet)

        if args.requested_arch is not None:
            self.requested_architecture_list = args.requested_arch.upper().split(",")
        else:
            self.requested_architecture_list = []

        if args.requested_target is not None:
            self.requested_target_list = args.requested_target.upper().split(",")
        else:
            self.requested_target_list = []

    def InputParametersConfiguredCallback(self) -> None:
        """Initializes the environment once input parameters are collected."""
        if (len(self.requested_package_list) == 0):
            self.requested_package_list = list(self.PlatformSettings.GetPackagesSupported())
        self.PlatformSettings.SetPackages(self.requested_package_list)

        if (len(self.requested_architecture_list) == 0):
            self.requested_architecture_list = list(self.PlatformSettings.GetArchitecturesSupported())
        self.PlatformSettings.SetArchitectures(self.requested_architecture_list)

        if (len(self.requested_target_list) == 0):
            self.requested_target_list = list(self.PlatformSettings.GetTargetsSupported())
        self.PlatformSettings.SetTargets(self.requested_target_list)



# @file Edk2Update
# Updates external dependencies for project_scope in workspace_path
# as listed in Platform Config file.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Updates external dependencies.

Contains a UpdateSettingsManager that must be subclassed in a build settings
file. This provides platform specific information to Edk2Update invocable
while allowing the invocable itself to remain platform agnostic.
"""
import argparse
import logging

from edk2toolext import edk2_logging
from edk2toolext.environment import self_describing_environment
from edk2toolext.invocables.edk2_multipkg_aware_invocable import (
    Edk2MultiPkgAwareInvocable,
    MultiPkgAwareSettingsInterface,
)


class UpdateSettingsManager(MultiPkgAwareSettingsInterface):
    """Platform specific settings for Edk2Update.

    Provides information necessary for `stuart_update.exe` or `edk2_update.py`
    when updating the platform.

    Update settings manager has no additional APIs not already defined in it's super class, however
    The class should still be overwritten by the platform.
    """


def build_env_changed(
        build_env: self_describing_environment.self_describing_environment,
        build_env_2: self_describing_environment.self_describing_environment
    ) -> bool:
    """Return True if build_env has changed."""
    return (build_env.paths != build_env_2.paths) or \
           (build_env.extdeps != build_env_2.extdeps) or \
           (build_env.plugins != build_env_2.plugins)


class Edk2Update(Edk2MultiPkgAwareInvocable):
    """Updates dependencies in workspace for active scopes."""

    MAX_RETRY_COUNT = 10

    def PerformUpdate(self) -> tuple:
        """Updates the dependencies."""
        ws_root = self.GetWorkspaceRoot()
        scopes = self.GetActiveScopes()
        skipped_dirs = self.GetSkippedDirectories()
        (build_env, shell_env) = self_describing_environment.BootstrapEnvironment(ws_root, scopes, skipped_dirs)
        (success, failure) = self_describing_environment.UpdateDependencies(ws_root, scopes, skipped_dirs)
        if success != 0:
            logging.log(edk2_logging.SECTION, f"\tUpdated/Verified {success} dependencies")
        return (build_env, shell_env, failure)

    def GetVerifyCheckRequired(self) -> bool:
        """Will not call self_describing_environment.VerifyEnvironment because ext_deps haven't been unpacked yet."""
        return False

    def GetSettingsClass(self) -> type:
        """Returns the UpdateSettingsManager class.

        !!! warning
            UpdateSettingsManager must be subclassed in your platform settings file.
        """
        return UpdateSettingsManager

    def GetLoggingFileName(self, loggerType: str) -> str:
        """Returns the filename (UPDATE_LOG) of where the logs for the Edk2CiBuild invocable are stored in."""
        return "UPDATE_LOG"

    def AddCommandLineOptions(self, parserObj: argparse.ArgumentParser) -> None:
        """Adds command line options to the argparser."""
        super().AddCommandLineOptions(parserObj)

    def RetrieveCommandLineOptions(self, args: argparse.Namespace) -> None:
        """Retrieve command line options from the argparser."""
        super().RetrieveCommandLineOptions(args)

    def Go(self) -> int:
        """Executes the core functionality of the Edk2Update invocable."""
        RetryCount = 0
        failure_count = 0
        logging.log(edk2_logging.SECTION, "Initial update of environment")

        (build_env_old, shell_env_old, _) = self.PerformUpdate()
        self_describing_environment.DestroyEnvironment()

        # Loop updating dependencies until there are 0 new dependencies or
        # we have exceeded retry count.  This allows dependencies to carry
        # files that influence the SDE.
        logging.log(edk2_logging.SECTION, "Second pass update of environment")
        while RetryCount < Edk2Update.MAX_RETRY_COUNT:
            (build_env, shell_env, failure_count) = self.PerformUpdate()

            if not build_env_changed(build_env, build_env_old):  # check if the environment changed on our last update
                break
            # if the environment has changed, increment the retry count and notify user
            RetryCount += 1
            logging.log(edk2_logging.SECTION,
                        f"Something in the environment changed. Updating environment again. Pass {RetryCount}")

            build_env_old = build_env
            self_describing_environment.DestroyEnvironment()

        if failure_count != 0:
            logging.error(f"We were unable to successfully update {failure_count} dependencies in environment")
        if RetryCount >= Edk2Update.MAX_RETRY_COUNT:
            logging.error(f"We did an update more than {Edk2Update.MAX_RETRY_COUNT} times.")
            logging.error("Please check your dependencies and make sure you don't have any circular ones.")
            return 1
        return failure_count


def main() -> None:
    """Entry point to invoke Edk2Update."""
    Edk2Update().Invoke()



# @file shell_environment.py
# This module contains code that helps to manage the build environment
# including PATH, PYTHONPATH, and ENV variables.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code that helps manage and maintain the build environment.

This management includes PATH, PYTHONPATH and ENV Variables.
"""
import copy
import logging
import os
import sys
from typing import Any

from edk2toolext.environment import var_dict

LOGGING_GROUP = "EnvDict"
MY_LOGGER = logging.getLogger(LOGGING_GROUP)


#
# Copy the Singleton pattern from...
#   https://stackoverflow.com/a/6798042
#
class Singleton(type): # noqa
    _instances = {}

    def __call__(cls, *args, **kwargs): # noqa
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class ShellEnvironment(metaclass=Singleton):
    """An active copy of the current OS environment.

    Allows for easy manipulation of the environment including taking
    screenshots (checkpoints) that are stored and can be accessed
    later.
    """
    # Easy definition for the very first checkpoint
    # when the environment is first created.
    INITIAL_CHECKPOINT = 0

    def __init__(self) -> None:
        """Inits the local environment with the initial os environment."""
        # Add all of our logging to the EnvDict group.
        self.logger = logging.getLogger(LOGGING_GROUP)

        # Initialize all other things.
        self.active_environ = None
        self.active_path = None
        self.active_pypath = None
        self.active_buildvars = var_dict.VarDict()
        self.checkpoints = []

        # Grab a copy of the environment as it exists.
        self.import_environment()

        # Create the initial checkpoint.
        self.checkpoint()

    #
    # Management methods.
    # These methods manage the singleton, the surrounding environment, and checkpoints.
    #
    def import_environment(self) -> None:
        """Loads the local environment with os environment."""
        # Create a complete copy of os.environ
        self.active_environ = dict()
        for key, value in os.environ.items():
            self.active_environ[key] = value

        # Record the PATH elements of the current environment.
        path = self.active_environ.get('PATH', "")

        # Filter removes empty elements.
        # List creates an actual list rather than a generator.
        # Filter removes empty strings.
        self.active_path = list(filter(None, path.split(os.pathsep)))

        # Record the PYTHONPATH elements of the current environment.
        # When reading PYTHONPATH, try reading the live path from sys.
        self.active_pypath = sys.path

        # Remove PATH and PYTHONPATH from environ copy to force use of active_path and active_pypath
        self.active_environ.pop("PATH", None)
        self.active_environ.pop("PYTHONPATH", None)

    def export_environment(self) -> None:
        """Exports enviornment to the OS."""
        # Purge all keys that aren't in the export.
        for key, value in os.environ.items():
            if key not in self.active_environ:
                os.environ.pop(key)

        # Export all internal keys.
        for key, value in self.active_environ.items():
            os.environ[key] = value

        # Set the PATH and PYTHONPATH vars.
        os.environ["PATH"] = os.pathsep.join(self.active_path)
        os.environ["PYTHONPATH"] = os.pathsep.join(self.active_pypath)

        sys.path = self.active_pypath

    def log_environment(self) -> None:
        """Logs the current environment to the logger."""
        self.logger.debug("FINAL PATH:")
        self.logger.debug(", ".join(self.active_path))

        self.logger.debug("FINAL PYTHONPATH:")
        self.logger.debug(", ".join(self.active_pypath))

        self.logger.debug("FINAL ENVIRON:")
        environ_list = []
        for key, value in self.active_environ.items():
            environ_list.append("({0}:{1})".format(key, value))
        self.logger.debug(", ".join(environ_list))

    def checkpoint(self) -> int:
        """Creates a checkpoint in time.

        Checkpoint stores the following:
        1. active_environment
        2. active_path
        3. active_pypath
        4. active_buildvars
        """
        new_index = len(self.checkpoints)
        self.checkpoints.append({
            'environ': copy.copy(self.active_environ),
            'path': self.active_path,
            'pypath': self.active_pypath,
            'buildvars': copy.copy(self.active_buildvars)
        })

        return new_index

    def restore_checkpoint(self, index: int) -> None:
        """Restore a specific checkpoint."""
        if index < len(self.checkpoints):
            check_point = self.checkpoints[index]
            self.active_environ = copy.copy(check_point['environ'])
            self.active_path = check_point['path']
            self.active_pypath = check_point['pypath']
            self.active_buildvars = copy.copy(check_point['buildvars'])

            self.export_environment()

        else:
            raise IndexError("Checkpoint %s does not exist" % index)

    def restore_initial_checkpoint(self) -> None:
        """Restore the initial checkpoint made."""
        self.restore_checkpoint(ShellEnvironment.INITIAL_CHECKPOINT)

    #
    # Environment manipulation methods.
    # These methods interact with the current environment.
    #
    def _internal_set_path(self, path_elements: str) -> None:
        self.active_path = list(path_elements)
        os.environ["PATH"] = os.pathsep.join(self.active_path)

    def _internal_set_pypath(self, path_elements: str) -> None:
        self.active_pypath = list(path_elements)
        os.environ["PYTHONPATH"] = os.pathsep.join(self.active_pypath)
        sys.path = self.active_pypath

    def set_path(self, new_path: str) -> None:
        """Set the path.

        Args:
            new_path (str): path to override with
        """
        self.logger.debug("Overriding PATH with new value.")
        if isinstance(new_path, str):
            new_path = list(new_path.split(os.pathsep))
        self._internal_set_path(new_path)

    def set_pypath(self, new_path: str) -> None:
        """Set the pypath.

        Args:
            new_path (str): path to override with
        """
        self.logger.debug("Overriding PYTHONPATH with new value.")
        if isinstance(new_path, str):
            new_path = list(new_path.split(os.pathsep))
        self._internal_set_pypath(new_path)

    def append_path(self, path_element: str) -> None:
        """Append to the end of path.

        if path_element already exists within path it will be removed
        from the current location and appended to the end

        Args:
            path_element (str): path element to append
        """
        self.logger.debug("Appending PATH element '%s'." % path_element)
        if path_element in self.active_path:
            # remove so we don't have duplicates but we respect the order
            # requested by the caller
            self.active_path.remove(path_element)
        self._internal_set_path(self.active_path + [path_element])

    def insert_path(self, path_element: str) -> None:
        """Insert at front of the path.

        if path_element already exists within path it will be removed
        from the current location and prepended to the front

        Args:
            path_element (str): path element to insert
        """
        self.logger.debug("Inserting PATH element '%s'." % path_element)
        if path_element in self.active_path:
            # remove so we don't have duplicates but we respect the order
            # requested by the caller
            self.active_path.remove(path_element)
        self._internal_set_path([path_element] + self.active_path)

    def append_pypath(self, path_element: str) -> None:
        """Append to the end of pypath.

        if path_element already exists within pypath it will be removed
        from the current location and appended to the end

        Args:
            path_element (str): path element to append
        """
        self.logger.debug("Appending PYTHONPATH element '%s'." % path_element)
        if path_element in self.active_pypath:
            # remove so we don't have duplicates but we respect the order
            # requested by the caller
            self.active_pypath.remove(path_element)
        self._internal_set_pypath(self.active_pypath + [path_element])

    def insert_pypath(self, path_element: str) -> None:
        """Insert at front of the pypath.

        if path_element already exists within pypath it will be removed
        from the current location and prepended to the front

        Args:
            path_element (str): path element to insert
        """
        self.logger.debug("Inserting PYTHONPATH element '%s'." % path_element)
        if path_element in self.active_pypath:
            # remove so we don't have duplicates but we respect the order
            # requested by the caller
            self.active_pypath.remove(path_element)
        self._internal_set_pypath([path_element] + self.active_pypath)

    def replace_path_element(self, old_path_element: str, new_path_element: str) -> None:
        """Replaces the PATH element.

        Generates a new PATH by iterating through the old PATH and replacing
        old_path_element with new_path_element where it is found.

        Args:
            old_path_element (str): element to replace
            new_path_element (str): element to replace with

        """
        self.logger.debug("Replacing PATH element {0} with {1}".format(old_path_element, new_path_element))
        self._internal_set_path([x if x != old_path_element else new_path_element for x in self.active_path])

    def replace_pypath_element(self, old_pypath_element: str, new_pypath_element: str) -> None:
        """Replaces the PYPATH element.

        Generates a new PYPATH by iterating through the old PYPATH and replacing
        old_pypath_element with new_pypath_element where it is found.

        Args:
            old_pypath_element (str): element to replace
            new_pypath_element (str): element to replace with

        """
        self.logger.debug("Replacing PYPATH element {0} with {1}".format(old_pypath_element, new_pypath_element))
        self._internal_set_pypath([x if x != old_pypath_element else new_pypath_element for x in self.active_pypath])

    def remove_path_element(self, path_element: str) -> None:
        """Removes the PATH element.

        Generates a new PATH by iterating through the old PATH and removing
        path_element if it is found.

        Args:
            path_element (str): path element to remove
        """
        self.logger.debug("Removing PATH element {0}".format(path_element))
        self._internal_set_path([x for x in self.active_path if x != path_element])

    def remove_pypath_element(self, pypath_element: str) -> None:
        """Removes the PYPATH element.

        Generates a new PYPATH by iterating through the old PYPATH and removing
        pypath_element if it is found.

        Args:
            pypath_element (str): pypath element to remove
        """
        self.logger.debug("Removing PYPATH element {0}".format(pypath_element))
        self._internal_set_pypath([x for x in self.active_pypath if x != pypath_element])

    def get_build_var(self, var_name: str) -> str:
        """Gets the build variable.

        Args:
            var_name (str): variable to get the value of
        Returns:
            (obj): value associated with the var_name
        """
        return self.active_buildvars.GetValue(var_name)

    def set_build_var(self, var_name: str, var_data: str) -> None:
        """Sets the variable as internal build variable.

        Unlike set_shell_var, var_data can be `None`; this sets var_name as a non-valued
        build variable (e.g. E1000_ENABLE). Additional information can be found at:
        https://www.tianocore.org/edk2-pytool-extensions/integrate/build/#setting-getting-environment-variables.

        !!! note
            Variables set in this manner are only accessable inside stuart, and are not an
            os environment variable. Refer to set_shell_var to set an os environment variable.

        Args:
            var_name (str): variable to set the value for
            var_data (obj): data to set
        """
        self.logger.debug(
            "Updating BUILD VAR element '%s': '%s'." % (var_name, var_data))
        self.active_buildvars.SetValue(var_name, var_data, '', overridable=True)

    def get_shell_var(self, var_name: str) -> str:
        """Gets the shell variable.

        Args:
            var_name (str): variable to get the value of

        Returns:
            (str): value associated with the var name
        """
        return self.active_environ.get(var_name, None)

    def set_shell_var(self, var_name: str, var_data: str) -> None:
        """Sets the variable as an OS environment variable.

        Args:
            var_name (str): variable to set the value for
            var_data (str): data to set

        Raises:
            (ValueError): var_data is None
        """
        # Check for the "special" shell vars.
        if var_data is None:
            raise ValueError("Unexpected var_data: None")
        if var_name.upper() == 'PATH':
            self.set_path(var_data)
        elif var_name.upper() == 'PYTHONPATH':
            self.set_pypath(var_data)
        else:
            self.logger.debug(
                "Updating SHELL VAR element '%s': '%s'." % (var_name, var_data))
            self.active_environ[var_name] = var_data
            os.environ[var_name] = var_data


def GetEnvironment() -> ShellEnvironment:
    """Returns the environment.

    Returns:
        (ShellEnvironment): Singleton class
    """
    return ShellEnvironment()


def GetBuildVars() -> var_dict.VarDict:
    """The current checkpoint buildvar values.

    Returns:
        (VarDict): A special dictionary containing build vars
    """
    #
    # Tricky!
    # Define a wrapper class that always forwards commands to the
    # BuildVars associated with the current environment.
    #
    # Will be deprecated.
    #
    class BuildVarsWrapper(object):
        def __init__(self) -> None:
            self.internal_shell_env = ShellEnvironment()

        def __getattr__(self, attrname: str) -> Any: # noqa: ANN401
            # Instead, invoke on the active BuildVars object.
            return getattr(self.internal_shell_env.active_buildvars, attrname)

    return BuildVarsWrapper()


#
# TODO: These are convenience methods that should be deprecated.
#
checkpoint_list = []


def CheckpointBuildVars() -> None:
    """Creates a checkpoint [a screenshot in time] of all current build var values."""
    global checkpoint_list
    new_checkpoint = ShellEnvironment().checkpoint()
    checkpoint_list.append(new_checkpoint)
    MY_LOGGER.debug("Created checkpoint {0} for build vars".format(new_checkpoint))


def RevertBuildVars() -> None:
    """Reverts all build var values to the most recent checkpoint."""
    global checkpoint_list
    if len(checkpoint_list) > 0:
        last_checkpoint = checkpoint_list.pop()
        MY_LOGGER.debug("Reverting to checkpoint {0} for build vars".format(last_checkpoint))
        ShellEnvironment().restore_checkpoint(last_checkpoint)
    else:
        MY_LOGGER.getLogger("No more checkpoints!")
        raise RuntimeError("No more checkpoints!")



# @file edk2_ci_setup.py
# Resolves all dependent repos for a CI environment.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code that supports CI/CD via the ci_setup invocable.

Contains a CISetupSettingsManager that must be subclassed in a build settings
file. This provides platform specific information to Edk2CiSetup invocable
while allowing the invocable itself to remain platform agnostic.
"""
import argparse
import logging
import os
from typing import Iterable

from edk2toolext.environment import repo_resolver
from edk2toolext.invocables.edk2_multipkg_aware_invocable import (
    Edk2MultiPkgAwareInvocable,
    MultiPkgAwareSettingsInterface,
)


class CiSetupSettingsManager(MultiPkgAwareSettingsInterface):
    """Platform specific settings for Edk2CiSetup.

    Provide information necessary for `stuart_ci_setup.exe` or
    `edk2_ci_setup.py` to successfully execute.

    !!! example: "Example of Overriding CiSetupSettingsManager"
        ```python
        from edk2toolext.invocables.edk2_ci_setup import CiSetupSettingsManager
        class CiManager(CiSetupSettingsManager):
            def GetDependencies(self):
                return {
                    "Path": "/Common/MU",
                    "Url":  "https://github.com/Microsoft/mu_tiano_plus.git"
                }
        ```
    """

    def GetDependencies(self) -> Iterable[dict]:
        """Get any Git Repository Dependendencies.

        This list of repositories will be resolved during the setup step.

        !!! tip
            Optional Override in subclass

        !!! tip
            Return an iterable of dictionary objects with the following fields
            ```json
            {
                Path: <required> Workspace relative path
                Url: <required> Url of git repo
                Commit: <optional> Commit to checkout of repo
                Branch: <optional> Branch to checkout (will checkout most recent commit in branch)
                Full: <optional> Boolean to do shallow or Full checkout.  (default is False)
                ReferencePath: <optional> Workspace relative path to git repo to use as "reference"
            }
            ```
        """
        return []


class Edk2CiBuildSetup(Edk2MultiPkgAwareInvocable):
    """Invocable supporting an iterative multi-package build and test process leveraging CI build plugins.

    Edk2CiBuildSetup sets up the necessary environment for Edk2CiBuild by preparing all necessary submodules.
    """
    def AddCommandLineOptions(self, parser: argparse.ArgumentParser) -> None:
        """Adds command line arguments to Edk2CiBuild."""
        parser.add_argument('-ignore', '--ignore-git', dest="git_ignore", action="store_true",
                            help="Whether to ignore errors in the git cloning process", default=False)
        parser.add_argument('--omnicache', '--reference', dest='omnicache_path',
                            default=os.environ.get('OMNICACHE_PATH'))
        parser.add_argument('-force', '--force-git', dest="git_force", action="store_true",
                            help="Whether to force git repos to clone in the git cloning process", default=False)
        parser.add_argument('-update-git', '--update-git', dest="git_update", action="store_true",
                            help="Whether to update git repos as needed in the git cloning process", default=False)
        super().AddCommandLineOptions(parser)

    def RetrieveCommandLineOptions(self, args: argparse.Namespace) -> None:
        """Retrieve command line options from the argparser."""
        self.git_ignore = args.git_ignore
        self.git_force = args.git_force
        self.git_update = args.git_update
        self.omnicache_path = args.omnicache_path
        if (self.omnicache_path is not None) and (not os.path.exists(self.omnicache_path)):
            logging.warning(f"Omnicache path set to invalid path: {args.omnicache_path}")
            self.omnicache_path = None

        super().RetrieveCommandLineOptions(args)

    def GetVerifyCheckRequired(self) -> bool:
        """Will not verify environment."""
        return False

    def GetSettingsClass(self) -> type:
        """Returns the CiSetupSettingsManager class.

        !!! warning
            CiSetupSettingsManager must be subclassed in your platform settings file.
        """
        return CiSetupSettingsManager

    def GetLoggingFileName(self, loggerType: str) -> str:
        """Returns the filename (CISETUP) of where the logs for the Edk2CiBuild invocable are stored in."""
        return "CISETUP"

    def Go(self) -> int:
        """Executes the core functionality of the Edk2CiSetup invocable."""
        setup_dependencies = self.PlatformSettings.GetDependencies()
        logging.debug(f"Dependencies list {setup_dependencies}")
        repos = repo_resolver.resolve_all(self.GetWorkspaceRoot(),
                                          setup_dependencies,
                                          ignore=self.git_ignore, force=self.git_force,
                                          update_ok=self.git_update, omnicache_dir=self.omnicache_path)

        logging.info(f"Repo resolver resolved {repos}")

        return 0 if None not in repos else -1


def main() -> None:
    """Entry point invoke Edk2CiBuild."""
    Edk2CiBuildSetup().Invoke()



# @file codeql.py
#
# Exports functions commonly needed for Stuart-based platforms to easily
# enable CodeQL in their platform build.
#
# Copyright (c) Microsoft Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""CodeQL Helper Functions.

Contains functions are intended to ease repo integration with CodeQL, ensure
consistent command-line usage across repos, and define standard scopes that
other plugins and tools can depend on for CodeQL operations.
"""
from argparse import ArgumentParser, Namespace
from typing import Tuple

from edk2toollib.utility_functions import GetHostInfo

from edk2toolext.environment.uefi_build import UefiBuilder


def add_command_line_option(parser: ArgumentParser) -> None:
    """Add the CodeQL command to the platform command line options.

    Args:
        parser (ArgumentParser): The argument parser used in this build.

    """
    parser.add_argument(
        '--codeql',
        dest='codeql',
        action='store_true',
        default=False,
        help="Optional - Produces CodeQL results from the build.")


def get_scopes(codeql_enabled: bool) -> Tuple[str]:
    """Return the active CodeQL scopes for this build.

    Args:
        codeql_enabled (bool): Whether CodeQL is enabled.

    Returns:
        Tuple[str]: A tuple of strings containing scopes that enable the
                    CodeQL plugin.
    """
    active_scopes = ()

    if codeql_enabled:
        if GetHostInfo().os == "Linux":
            active_scopes += ("codeql-linux-ext-dep",)
        else:
            active_scopes += ("codeql-windows-ext-dep",)
        active_scopes += ("codeql-build", "codeql-analyze")

    return active_scopes


def is_codeql_enabled_on_command_line(args: Namespace) -> bool:
    """Return whether CodeQL was enabled on the command line.

    Args:
        args (Namespace): Object holding a string representation of command
                          line arguments.

    Returns:
        bool: True if CodeQL is enabled on the command line. Otherwise, false.
    """
    return args.codeql


def set_audit_only_mode(uefi_builder: UefiBuilder) -> None:
    """Configure the CodeQL plugin to run in audit only mode.

    Args:
        uefi_builder (UefiBuilder): The UefiBuilder object for this platform
                                    build.

    """
    uefi_builder.env.SetValue(
        "STUART_CODEQL_AUDIT_ONLY",
        "true",
        "Platform Defined")
